"""
this code is from pyart.core.transforms by Helmus, J.J. & Collis, S.M.
# https://github.com/ARM-DOE/pyart
=====================

Transformations between coordinate systems. Routines for converting between
Cartesian/Cartographic (x, y, z), Geographic (latitude, longitude, altitude)
and antenna (azimuth, elevation, range) coordinate systems.

.. autosummary::
    :toctree: generated/

    antenna_to_cartesian
    antenna_vectors_to_cartesian
    antenna_to_cartesian_track_relative
    antenna_to_cartesian_earth_relative
    antenna_to_cartesian_aircraft_relative

    cartesian_to_geographic
    cartesian_vectors_to_geographic
    geographic_to_cartesian
    cartesian_to_geographic_aeqd
    geographic_to_cartesian_aeqd

    _interpolate_axes_edges
    _interpolate_azimuth_edges
    _interpolate_elevation_edges
    _interpolate_range_edges
    _half_angle_complex


"""

# the code for transforms in this file were adapted from pyart by Helmus, J.J. & Collis, S.M.
# https://github.com/ARM-DOE/pyart

import warnings

import numpy as np
import xarray as xr
try:
    import pyproj
    _PYPROJ_AVAILABLE = True
except ImportError:
    _PYPROJ_AVAILABLE = False

PI = np.pi
DEFAULT_EFFECTIVE_EARTH_RADIUS = 6371.0 * 1000.0 * 4.0 / 3.0


def resolve_effective_earth_radius(effective_earth_radius=None):
    """Return a validated effective earth radius in meters."""
    if effective_earth_radius is None:
        return DEFAULT_EFFECTIVE_EARTH_RADIUS
    radius = float(effective_earth_radius)
    if radius <= 0:
        raise ValueError("effective_earth_radius must be positive")
    return radius


def _deg2rad_sin_cos(values):
    """Return sine and cosine of degree-valued angles."""
    radians = np.deg2rad(np.asarray(values))
    return np.sin(radians), np.cos(radians)


def _antenna_vectors_to_cartesian_impl(
    ranges,
    azimuths,
    elevations,
    h=0.0,
    effective_earth_radius=None,
):
    """Vectorized high-precision antenna-to-Cartesian transform."""
    ranges = np.asarray(ranges)
    azimuths = np.asarray(azimuths)
    elevations = np.asarray(elevations)

    r = ranges[np.newaxis, :]
    sin_a, cos_a = _deg2rad_sin_cos(azimuths)
    sin_e, cos_e = _deg2rad_sin_cos(elevations)
    sin_a = sin_a[:, np.newaxis]
    cos_a = cos_a[:, np.newaxis]
    sin_e = sin_e[:, np.newaxis]
    cos_e = cos_e[:, np.newaxis]

    R = resolve_effective_earth_radius(effective_earth_radius)
    radial = r * cos_e
    gate_radius = np.sqrt(radial * radial + (R + h + r * sin_e) ** 2)
    z = gate_radius - R
    arc_arg = np.clip(radial / gate_radius, -1.0, 1.0)
    s = R * np.arcsin(arc_arg)
    x = s * sin_a
    y = s * cos_a
    return x, y, z


def antenna_to_cartesian(ranges, azimuths, elevations, effective_earth_radius=None):
    """
    Return Cartesian coordinates from antenna coordinates.

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in meters.
    azimuths : array
        Azimuth angle of the radar in degrees.
    elevations : array
        Elevation angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    The calculation for Cartesian coordinate is adapted from equations
    2.28(b) and 2.28(c) of Doviak and Zrnic [1]_ assuming a
    standard atmosphere (4/3 Earth's radius model).

    .. math::

        z = \\sqrt{r^2+R^2+2*r*R*sin(\\theta_e)} - R

        s = R * arcsin(\\frac{r*cos(\\theta_e)}{R+z})

        x = s * sin(\\theta_a)

        y = s * cos(\\theta_a)

    Where r is the distance from the radar to the center of the gate,
    :math:`\\theta_a` is the azimuth angle, :math:`\\theta_e` is the
    elevation angle, s is the arc length, and R is the effective radius
    of the earth, taken to be 4/3 the mean radius of earth (6371 km).

    References
    ----------
    .. [1] Doviak and Zrnic, Doppler Radar and Weather Observations, Second
        Edition, 1993, p. 21.

    """
    return antenna_to_cartesian_cwr(
        ranges,
        azimuths,
        elevations,
        0.0,
        effective_earth_radius=effective_earth_radius,
    )

def antenna_to_cartesian_cwr(ranges, azimuths, elevations, h, effective_earth_radius=None):
    """
        Return Cartesian coordinates from antenna coordinates.
        Parameters
        ----------
        ranges : array
            Distances to the center of the radar gates (bins) in meters.
        azimuths : array
            Azimuth angle of the radar in degrees.
        elevations : array
            Elevation angle of the radar in degrees.
        h : constant
            Altitude of the instrument, above sea level, units:m.
        Returns
        -------
        x, y, z : array
            Cartesian coordinates in meters from the radar.
        Notes
        -----
        The calculation for Cartesian coordinate is adapted from equations
        2.28(b) and 2.28(c) of Doviak and Zrnic [1]_ assuming a
        standard atmosphere (4/3 Earth's radius model).
        .. math::
            z = \\sqrt{r^2+R^2+2*r*R*sin(\\theta_e)} - R
            s = R * arcsin(\\frac{r*cos(\\theta_e)}{R+z})
            x = s * sin(\\theta_a)
            y = s * cos(\\theta_a)
        Where r is the distance from the radar to the center of the gate,
        :math:`\\theta_a` is the azimuth angle, :math:`\\theta_e` is the
        elevation angle, s is the arc length, and R is the effective radius
        of the earth, taken to be 4/3 the mean radius of earth (6371 km).
        References
        ----------
        .. [1] Doviak and Zrnic, Doppler Radar and Weather Observations, Second
            Edition, 1993, p. 21.
    """
    sin_e, cos_e = _deg2rad_sin_cos(elevations)
    sin_a, cos_a = _deg2rad_sin_cos(azimuths)
    R = resolve_effective_earth_radius(effective_earth_radius)
    r = np.asarray(ranges) * 1.0  # distances to gates in meters.

    radial = r * cos_e
    gate_radius = np.sqrt(radial ** 2 + (R + h + r * sin_e) ** 2)
    z = gate_radius - R
    s = R * np.arcsin(np.clip(radial / gate_radius, -1.0, 1.0))  # arc length in m.
    x = s * sin_a
    y = s * cos_a

    return x, y, z

def cartesian_xyz_to_antenna(x, y, z, h, effective_earth_radius=None):
    """
    Invert Cartesian coordinates back to antenna coordinates.
    x, units: meters
    y, units: meters
    z, units: meters
    h, units: meters
    return azimuth, ranges, elevation
    """
    R = resolve_effective_earth_radius(effective_earth_radius)
    s = np.hypot(x, y)
    radar_radius = R + h
    gate_radius = R + z
    ranges_sq = radar_radius ** 2 + gate_radius ** 2 - 2 * radar_radius * gate_radius * np.cos(s / R)
    ranges = np.sqrt(np.maximum(ranges_sq, 0.0))
    with np.errstate(invalid="ignore", divide="ignore"):
        cos_arg = (radar_radius ** 2 + ranges ** 2 - gate_radius ** 2) / (2 * radar_radius * ranges)
    cos_arg = np.clip(cos_arg, -1.0, 1.0)
    elevation = (np.arccos(cos_arg) - np.pi / 2) * 180. / np.pi

    azimuth = _azimuth(x, y)
    return azimuth, ranges, elevation

def cartesian_xy_elevation_to_range_z(x, y, elevation, h, effective_earth_radius=None):
    """Invert planar coordinates and elevation back to slant range and height.

    Given planar distance ``s = sqrt(x^2 + y^2)`` and elevation ``el``, this
    uses the same 4/3-earth-radius model as :func:`antenna_to_cartesian_cwr`.
    """
    R = resolve_effective_earth_radius(effective_earth_radius)
    s = np.hypot(x, y)
    phi = s / R
    el = np.deg2rad(elevation)
    denom = np.cos(phi) - np.sin(phi) * np.tan(el)
    radar_radius = R + h
    with np.errstate(invalid="ignore", divide="ignore"):
        Rp = np.divide(radar_radius, denom)
    z = Rp - R
    ranges_sq = radar_radius ** 2 + (R + z) ** 2 - 2 * radar_radius * (R + z) * np.cos(phi)
    ranges = np.sqrt(np.maximum(ranges_sq, 0.0))
    az = _azimuth(x, y)
    return az, ranges, z


cartesian_to_antenna_cwr = cartesian_xy_elevation_to_range_z

def _azimuth(x, y):
    """Compute azimuth from Cartesian x/y offsets relative to the radar."""
    az = np.mod(np.pi / 2 - np.arctan2(y, x), 2 * np.pi)
    return np.rad2deg(az)

def antenna_vectors_to_cartesian_cwr(ranges, azimuths, elevations, h=0, edges=False,
                                     effective_earth_radius=None):
    """
    Calculate Cartesian coordinate for gates from antenna coordinate vectors.
    Calculates the Cartesian coordinates for the gate centers or edges for
    all gates from antenna coordinate vectors assuming a standard atmosphere
    (4/3 Earth's radius model). See :py:func:`pyart.util.antenna_to_cartesian`
    for details.
    Parameters
    ----------
    ranges : array, 1D.
        Distances to the center of the radar gates (bins) in meters.
    azimuths : array, 1D.
        Azimuth angles of the rays in degrees.
    elevations : array, 1D.
        Elevation angles of the rays in degrees.
    edges : bool, optional
        True to calculate the coordinates of the gate edges by interpolating
        between gates and extrapolating at the boundaries.  False to
        calculate the gate centers.
    Returns
    -------
    x, y, z : array, 2D
        Cartesian coordinates in meters from the center of the radar to the
        gate centers or edges.
    """
    if edges:
        if len(ranges) != 1:
            ranges = _interpolate_range_edges(ranges)
        if len(elevations) != 1:
            elevations = _interpolate_elevation_edges(elevations)
        if len(azimuths) != 1:
            azimuths = _interpolate_azimuth_edges(azimuths)
    return _antenna_vectors_to_cartesian_impl(
        ranges,
        azimuths,
        elevations,
        h,
        effective_earth_radius=effective_earth_radius,
    )


def antenna_vectors_to_cartesian(ranges, azimuths, elevations, edges=False,
                                 effective_earth_radius=None):
    """
    Calculate Cartesian coordinate for gates from antenna coordinate vectors.

    Calculates the Cartesian coordinates for the gate centers or edges for
    all gates from antenna coordinate vectors assuming a standard atmosphere
    (4/3 Earth's radius model). See :py:func:`pyart.util.antenna_to_cartesian`
    for details.

    Parameters
    ----------
    ranges : array, 1D.
        Distances to the center of the radar gates (bins) in meters.
    azimuths : array, 1D.
        Azimuth angles of the rays in degrees.
    elevations : array, 1D.
        Elevation angles of the rays in degrees.
    edges : bool, optional
        True to calculate the coordinates of the gate edges by interpolating
        between gates and extrapolating at the boundaries. False to
        calculate the gate centers.

    Returns
    -------
    x, y, z : array, 2D
        Cartesian coordinates in meters from the center of the radar to the
        gate centers or edges.

    """
    assert isinstance(ranges, (np.ndarray, xr.DataArray)), "check input dtype!"
    assert isinstance(azimuths, (np.ndarray, xr.DataArray)), "check input dtype!"
    assert isinstance(elevations, (np.ndarray, xr.DataArray)), "check input dtype!"
    if edges:
        if len(ranges) != 1:
            ranges = _interpolate_range_edges(ranges)
        if len(elevations) != 1:
            elevations = _interpolate_elevation_edges(elevations)
        if len(azimuths) != 1:
            azimuths = _interpolate_azimuth_edges(azimuths)
    return antenna_vectors_to_cartesian_cwr(
        ranges,
        azimuths,
        elevations,
        h=0.0,
        edges=edges,
        effective_earth_radius=effective_earth_radius,
    )

def antenna_vectors_to_cartesian_rhi(ranges, azimuths, elevations, h, BeamWidth=1,
                                     effective_earth_radius=None):
    """
    Build an RHI Cartesian coordinate system while accounting for beam width.
    :param ranges: gate-center ranges.
    :param azimuths: sweep azimuth.
    :param elevations: sweep elevation samples.
    :param BeamWidth: beam width in degrees.
    :param h: radar altitude.
    """
    assert elevations.size == 1, "not rhi, check!"
    assert azimuths.size == 1, "not rhi, check!"
    ranges = np.asarray(ranges)
    azimuths = np.asarray(azimuths)
    elevs = np.asarray([elevations - BeamWidth / 2.0, elevations + BeamWidth / 2.0]).reshape(-1)
    rg = np.broadcast_to(ranges, (elevs.size, ranges.size))
    azg = np.broadcast_to(azimuths.reshape(1, -1)[0:1], (elevs.size, ranges.size))
    eleg = np.broadcast_to(elevs[:, None], rg.shape)
    return antenna_to_cartesian_cwr(
        rg,
        azg,
        eleg,
        h,
        effective_earth_radius=effective_earth_radius,
    )

def antenna_vectors_to_cartesian_vcs(ranges, azimuths, elevations, h, BeamWidth=1,
                                     effective_earth_radius=None):
    """
    Build a vertical cross-section coordinate system while accounting for beam width.
    :param ranges: gate-center ranges.
    :param azimuths: azimuth samples.
    :param elevations: elevation samples.
    :param BeamWidth: beam width in degrees.
    :param h: radar altitude.
    """
    assert elevations.ndim == 1, "check ,may input data is not right!"
    assert azimuths.ndim == 1, "check ,may input data is not right!"
    assert ranges.ndim == 1, "check ,may input data is not right!"

    eleg = np.stack([elevations-BeamWidth/2., elevations+BeamWidth/2.], axis=0)
    rg = np.stack([ranges, ranges], axis=0)
    azg = np.stack([azimuths, azimuths], axis=0)
    return antenna_to_cartesian_cwr(
        rg,
        azg,
        eleg,
        h,
        effective_earth_radius=effective_earth_radius,
    )


def _interpolate_range_edges(ranges):
    """ Interpolate the edges of the range gates from their centers. """
    edges = np.empty((ranges.shape[0] + 1, ), dtype=ranges.dtype)
    edges[1:-1] = (ranges[:-1] + ranges[1:]) / 2.
    edges[0] = ranges[0] - (ranges[1] - ranges[0]) / 2.
    edges[-1] = ranges[-1] + (ranges[-1] - ranges[-2]) / 2.
    edges[edges < 0] = 0    # do not allow range to become negative
    return edges


def _interpolate_elevation_edges(elevations):
    """ Interpolate the edges of the elevation angles from their centers. """
    edges = np.empty((elevations.shape[0]+1, ), dtype=elevations.dtype)
    edges[1:-1] = (elevations[:-1] + elevations[1:]) / 2.
    edges[0] = elevations[0] - (elevations[1] - elevations[0]) / 2.
    edges[-1] = elevations[-1] + (elevations[-1] - elevations[-2]) / 2.
    edges[edges > 180] = 180.   # prevent angles from going below horizon
    edges[edges < 0] = 0.
    return edges


def _interpolate_azimuth_edges(azimuths):
    """ Interpolate the edges of the azimuth angles from their centers. """
    edges = np.empty((azimuths.shape[0]+1, ), dtype=azimuths.dtype)
    # perform interpolation and extrapolation in complex plane to
    # account for periodic nature of azimuth angle.
    azimuths = np.exp(1.j*np.deg2rad(azimuths))

    edges[1:-1] = np.angle(azimuths[1:] + azimuths[:-1], deg=True)

    half_angle = _half_angle_complex(azimuths[0], azimuths[1])
    edges[0] = (np.angle(azimuths[0], deg=True) - half_angle) % 360.

    half_angle = _half_angle_complex(azimuths[-1], azimuths[-2])
    edges[-1] = (np.angle(azimuths[-1], deg=True) + half_angle) % 360.

    edges[edges < 0] += 360     # range from [-180, 180] to [0, 360]
    return edges


def _half_angle_complex(complex_angle1, complex_angle2):
    """
    Return half the angle between complex numbers on the unit circle.

    Parameters
    ----------
    complex_angle1, complex_angle2 : complex
        Complex numbers representing unit vectors on the unit circle

    Returns
    -------
    half_angle : float
        Half the angle between the unit vectors in degrees.

    """
    dot_product = np.real(complex_angle1 * np.conj(complex_angle2))
    if dot_product > 1:
        warnings.warn("dot_product is larger than one.")
        dot_product = 1.
    full_angle_rad = np.arccos(dot_product)
    half_angle_rad = full_angle_rad / 2.
    half_angle_deg = np.rad2deg(half_angle_rad)
    return half_angle_deg


def _interpolate_axes_edges(axes):
    """ Interpolate the edges of the axes gates from their centers. """
    edges = np.empty((axes.shape[0] + 1, ), dtype=axes.dtype)
    edges[1:-1] = (axes[:-1] + axes[1:]) / 2.
    edges[0] = axes[0] - (axes[1] - axes[0]) / 2.
    edges[-1] = axes[-1] - (axes[-2] - axes[-1]) / 2.
    return edges


def antenna_to_cartesian_track_relative(ranges, rot, roll, drift, tilt, pitch):
    """
    Calculate track-relative Cartesian coordinates from radar coordinates.

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    roll : array
        Roll angle of the radar in degrees.
    drift : array
        Drift angle of the radar in degrees.
    tilt : array
        Tilt angle of the radar in degrees.
    pitch : array
        Pitch angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    track-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    roll = np.radians(roll)             # roll angle in radians.
    drift = np.radians(drift)           # drift angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    pitch = np.radians(pitch)           # pitch angle in radians.
    r = ranges * 1000.0                 # distances to gates in meters.

    x = r * (np.cos(rot + roll) * np.sin(drift) * np.cos(tilt) *
             np.sin(pitch) + np.cos(drift) * np.sin(rot + roll) *
             np.cos(tilt) - np.sin(drift) * np.cos(pitch) * np.sin(tilt))
    y = r * (-1. * np.cos(rot + roll) * np.cos(drift) * np.cos(tilt) *
             np.sin(pitch) + np.sin(drift) * np.sin(rot + roll) *
             np.cos(tilt) + np.cos(drift) * np.cos(pitch) * np.sin(tilt))
    z = r * (np.cos(pitch) * np.cos(tilt) * np.cos(rot + roll) +
             np.sin(pitch) * np.sin(tilt))
    return x, y, z


def antenna_to_cartesian_earth_relative(
        ranges, rot, roll, heading, tilt, pitch):
    """
    Calculate earth-relative Cartesian coordinates from radar coordinates

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    roll : array
        Roll angle of the radar in degrees.
    heading : array
        Heading (compass) angle of the radar in degrees clockwise from north.
    tilt : array
        Tilt angle of the radar in degrees.
    pitch : array
        Pitch angle of the radar in degrees.

    Returns
    -------
    x, y, z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    earth-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    roll = np.radians(roll)             # roll angle in radians.
    heading = np.radians(heading)       # drift angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    pitch = np.radians(pitch)           # pitch angle in radians.
    r = ranges * 1000.0                 # distances to gates in meters.

    x = r * (-1. * np.cos(rot + roll) * np.sin(heading) * np.cos(tilt) *
             np.sin(pitch) + np.cos(heading) * np.sin(rot + roll) *
             np.cos(tilt) + np.sin(heading) * np.cos(pitch) * np.sin(tilt))
    y = r * (-1. * np.cos(rot + roll) * np.cos(heading) * np.cos(tilt) *
             np.sin(pitch) - np.sin(heading) * np.sin(rot + roll) *
             np.cos(tilt) + np.cos(heading) * np.cos(pitch) * np.sin(tilt))
    z = r * (np.cos(pitch) * np.cos(tilt) * np.cos(rot + roll) +
             np.sin(pitch) * np.sin(tilt))
    return x, y, z


def antenna_to_cartesian_aircraft_relative(ranges, rot, tilt):
    """
    Calculate aircraft-relative Cartesian coordinates from radar coordinates.

    Parameters
    ----------
    ranges : array
        Distances to the center of the radar gates (bins) in kilometers.
    rot : array
        Rotation angle of the radar in degrees.
    tilt : array
        Tilt angle of the radar in degrees.

    Returns
    -------
    X, Y, Z : array
        Cartesian coordinates in meters from the radar.

    Notes
    -----
    Project native (polar) coordinate radar sweep data onto
    earth-relative Cartesian coordinate grid.

    References
    ----------
    .. [1] Lee et al. (1994) Journal of Atmospheric and Oceanic Technology.

    """
    rot = np.radians(rot)               # rotation angle in radians.
    tilt = np.radians(tilt)             # tilt angle in radians.
    r = ranges * 1000.0                 # distances to gates in meters.
    x = r * np.cos(tilt) * np.sin(rot)
    y = r * np.sin(tilt)
    z = r * np.cos(rot) * np.cos(tilt)
    return x, y, z


def geographic_to_cartesian(lon, lat, projparams):
    """
    Geographic to Cartesian coordinate transform.

    Transform a set of Geographic coordinate (lat, lon) to a
    Cartesian/Cartographic coordinate (x, y) using pyproj or a build in
    Azimuthal equidistant projection.

    Parameters
    ----------
    lon, lat : array-like
        Geographic coordinates in degrees.
    projparams : dict or str
        Projection parameters passed to pyproj.Proj. If this parameter is a
        dictionary with a 'proj' key equal to 'pyart_aeqd' then a azimuthal
        equidistant projection will be used that is native to Py-ART and
        does not require pyproj to be installed. In this case a non-default
        value of R can be specified by setting the 'R' key to the desired
        value.

    Returns
    -------
    x, y : array-like
        Cartesian coordinates in meters unless projparams defines a value for R
        in different units.

    """
    if isinstance(projparams, dict) and projparams.get('proj') == 'pyart_aeqd':
        # Use Py-ART's Azimuthal equidistance projection
        lon_0 = projparams['lon_0']
        lat_0 = projparams['lat_0']
        if 'R' in projparams:
            R = projparams['R']
            x, y = geographic_to_cartesian_aeqd(lon, lat, lon_0, lat_0, R)
        else:
            x, y = geographic_to_cartesian_aeqd(lon, lat, lon_0, lat_0)
    else:
        # Use pyproj for the projection
        # check that pyproj is available
        if not _PYPROJ_AVAILABLE:
            raise Exception(
                "PyProj is required to use geographic_to_cartesian "
                "with a projection other than pyart_aeqd but it is not "
                "installed")
        proj = pyproj.Proj(projparams)
        x, y = proj(lon, lat, inverse=False)
    return x, y


def geographic_to_cartesian_aeqd(lon, lat, lon_0, lat_0, R=6370997.):
    """
    Azimuthal equidistant geographic to Cartesian coordinate transform.

    Transform a set of geographic coordinates (lat, lon) to
    Cartesian/Cartographic coordinates (x, y) using a azimuthal equidistant
    map projection [1].

    .. math::

        x = R * k * \\cos(lat) * \\sin(lon - lon_0)

        y = R * k * [\\cos(lat_0) * \\sin(lat) -
                     \\sin(lat_0) * \\cos(lat) * \\cos(lon - lon_0)]

        k = c / \\sin(c)

        c = \\arccos(\\sin(lat_0) * \\sin(lat) +
                     \\cos(lat_0) * \\cos(lat) * \\cos(lon - lon_0))

    Where x, y are the Cartesian position from the center of projection;
    lat, lon the corresponding latitude and longitude; lat_0, lon_0 are the
    latitude and longitude of the center of the projection; R is the radius of
    the earth (defaults to ~6371 km).

    Parameters
    ----------
    lon, lat : array-like
        Longitude and latitude coordinates in degrees.
    lon_0, lat_0 : float
        Longitude and latitude, in degrees, of the center of the projection.
    R : float, optional
        Earth radius in the same units as x and y. The default value is in
        units of meters.

    Returns
    -------
    x, y : array
        Cartesian coordinates in the same units as R, typically meters.

    References
    ----------
    .. [1] Snyder, J. P. Map Projections--A Working Manual. U. S. Geological
        Survey Professional Paper 1395, 1987, pp. 191-202.

    """
    lon = np.asarray(lon)
    lat = np.asarray(lat)

    lon_rad = np.deg2rad(lon)
    lat_rad = np.deg2rad(lat)

    lat_0_rad = np.deg2rad(lat_0)
    lon_0_rad = np.deg2rad(lon_0)
    sin_lat_0 = np.sin(lat_0_rad)
    cos_lat_0 = np.cos(lat_0_rad)
    sin_lat = np.sin(lat_rad)
    cos_lat = np.cos(lat_rad)

    lon_diff_rad = lon_rad - lon_0_rad

    arg_arccos = np.clip(
        sin_lat_0 * sin_lat + cos_lat_0 * cos_lat * np.cos(lon_diff_rad),
        -1.0,
        1.0,
    )
    c = np.arccos(arg_arccos)
    sin_c = np.sin(c)
    k = np.divide(c, sin_c, out=np.ones_like(c, dtype=np.result_type(c, float)), where=sin_c != 0)

    x = R * k * cos_lat * np.sin(lon_diff_rad)
    y = R * k * (cos_lat_0 * sin_lat - sin_lat_0 * cos_lat * np.cos(lon_diff_rad))
    return x, y


def cartesian_to_geographic(x, y, projparams):
    """
    Cartesian to Geographic coordinate transform.

    Transform a set of Cartesian/Cartographic coordinates (x, y) to a
    geographic coordinate system (lat, lon) using pyproj or a build in
    Azimuthal equidistant projection.

    Parameters
    ----------
    x, y : array-like
        Cartesian coordinates in meters unless R is defined in different units
        in the projparams parameter.
    projparams : dict or str
        Projection parameters passed to pyproj.Proj. If this parameter is a
        dictionary with a 'proj' key equal to 'pyart_aeqd' then a azimuthal
        equidistant projection will be used that is native to Py-ART and
        does not require pyproj to be installed. In this case a non-default
        value of R can be specified by setting the 'R' key to the desired
        value.

    Returns
    -------
    lon, lat : array
        Longitude and latitude of the Cartesian coordinates in degrees.

    """
    if isinstance(projparams, dict) and projparams.get('proj') == 'pyart_aeqd':
        # Use Py-ART's Azimuthal equidistance projection
        lon_0 = projparams['lon_0']
        lat_0 = projparams['lat_0']
        if 'R' in projparams:
            R = projparams['R']
            lon, lat = cartesian_to_geographic_aeqd(x, y, lon_0, lat_0, R)
        else:
            lon, lat = cartesian_to_geographic_aeqd(x, y, lon_0, lat_0)
    else:
        # Use pyproj for the projection
        # check that pyproj is available
        if not _PYPROJ_AVAILABLE:
            raise Exception(
                "PyProj is required to use cartesian_to_geographic "
                "with a projection other than pyart_aeqd but it is not "
                "installed")
        proj = pyproj.Proj(projparams)
        lon, lat = proj(x, y, inverse=True)
    return lon, lat


def cartesian_vectors_to_geographic(x, y, projparams, edges=False):
    """
    Cartesian vectors to Geographic coordinate transform.

    Transform a set of Cartesian/Cartographic coordinate vectors (x, y) to a
    geographic coordinate system (lat, lon) using pyproj or a build in
    Azimuthal equidistant projection finding the coordinates edges in
    Cartesian space if requested.

    Parameters
    ----------
    x, y : array 1D.
        Cartesian coordinate vectors in meters unless R is defined in
        different units in the projparams parameter.
    projparams : dict or str
        Projection parameters passed to pyproj.Proj. If this parameter is a
        dictionary with a 'proj' key equal to 'pyart_aeqd' then a azimuthal
        equidistant projection will be used that is native to Py-ART and
        does not require pyproj to be installed. In this case a
        non-default value of R can be specified by setting the 'R' key to the
        desired value.
    edges : bool, optional
        True to calculate the coordinates of the geographic edges by
        interpolating between Cartesian points and extrapolating at the
        boundaries. False to calculate the coordinate centers.

    Returns
    -------
    lon, lat : array
        Longitude and latitude of the Cartesian coordinates in degrees.

    """
    if edges:
        if len(x) > 1:
            x = _interpolate_axes_edges(x)
        if len(y) > 1:
            y = _interpolate_axes_edges(y)
    x, y = np.meshgrid(x, y)
    return cartesian_to_geographic(x, y, projparams)


def cartesian_to_geographic_aeqd(x, y, lon_0, lat_0, R=6370997.):
    """
    Azimuthal equidistant Cartesian to geographic coordinate transform.

    Transform a set of Cartesian/Cartographic coordinates (x, y) to
    geographic coordinate system (lat, lon) using a azimuthal equidistant
    map projection [1].

    .. math::

        lat = \\arcsin(\\cos(c) * \\sin(lat_0) +
                       (y * \\sin(c) * \\cos(lat_0) / \\rho))

        lon = lon_0 + \\arctan2(
            x * \\sin(c),
            \\rho * \\cos(lat_0) * \\cos(c) - y * \\sin(lat_0) * \\sin(c))

        \\rho = \\sqrt(x^2 + y^2)

        c = \\rho / R

    Where x, y are the Cartesian position from the center of projection;
    lat, lon the corresponding latitude and longitude; lat_0, lon_0 are the
    latitude and longitude of the center of the projection; R is the radius of
    the earth (defaults to ~6371 km). lon is adjusted to be between -180 and
    180.

    Parameters
    ----------
    x, y : array-like
        Cartesian coordinates in the same units as R, typically meters.
    lon_0, lat_0 : float
        Longitude and latitude, in degrees, of the center of the projection.
    R : float, optional
        Earth radius in the same units as x and y. The default value is in
        units of meters.

    Returns
    -------
    lon, lat : array
        Longitude and latitude of Cartesian coordinates in degrees.

    References
    ----------
    .. [1] Snyder, J. P. Map Projections--A Working Manual. U. S. Geological
        Survey Professional Paper 1395, 1987, pp. 191-202.

    """
    x = np.asarray(x)
    y = np.asarray(y)

    lat_0_rad = np.deg2rad(lat_0)
    lon_0_rad = np.deg2rad(lon_0)
    sin_lat_0 = np.sin(lat_0_rad)
    cos_lat_0 = np.cos(lat_0_rad)

    rho = np.hypot(x, y)
    c = rho / R
    sin_c = np.sin(c)
    cos_c = np.cos(c)
    with np.errstate(invalid="ignore", divide="ignore"):
        arcsin_arg = cos_c * sin_lat_0 + np.divide(y * sin_c * cos_lat_0, rho)
    lat_rad = np.arcsin(np.clip(arcsin_arg, -1.0, 1.0))
    lat_rad = np.where(rho == 0, lat_0_rad, lat_rad)
    lat_deg = np.rad2deg(lat_rad)

    x1 = x * sin_c
    x2 = rho * cos_lat_0 * cos_c - y * sin_lat_0 * sin_c
    lon_rad = lon_0_rad + np.arctan2(x1, x2)
    lon_deg = np.rad2deg(lon_rad)
    lon_deg = ((lon_deg + 180.0) % 360.0) - 180.0

    return lon_deg, lat_deg



# -*- coding: utf-8 -*-
"""Vertical section plotting helpers with explicit coordinate handling."""

import matplotlib.pyplot as plt

from ._plot_core import (
    ColorbarOptions,
    PlotStyle,
    as_numpy,
    create_standard_axes,
    default_sweep_title,
    ensure_ppi_scan,
    geographic_line_to_cartesian,
    interpolate_points_along_line,
    lonlat_section_points,
    plot_vertical_section,
    resolve_field_data,
    resolve_field_style,
)


class VerticalSection(object):
    """Backward-compatible RHI and VCS plotting helper."""

    def __init__(self, radar=None, **legacy_kwargs):
        legacy_radar = legacy_kwargs.pop("NuistRadar", None)
        if legacy_kwargs:
            raise TypeError("unexpected keyword arguments: %s" % sorted(legacy_kwargs))
        self.NRadar = radar if radar is not None else legacy_radar

    def RHI(
        self,
        azimuth,
        field_name,
        range_mode=None,
        height=(0, 18),
        title=None,
        clabel=None,
        continuously=False,
        normvar=None,
        cmap=None,
        cmap_bins=None,
        colorbar_orientation="horizontal",
        colorbar_ticks=None,
        colorbar_ticklabels=None,
        axis_labels=None,
        **kwargs
    ):
        section = self.NRadar.extract_rhi(azimuth=azimuth, field_name=field_name, range_mode=range_mode)
        field, field_key = resolve_field_data(self.NRadar, 0, field_name, range_mode=range_mode)
        style = resolve_field_style(
            self.NRadar,
            0,
            field_key,
            field,
            value_range=normvar,
            cmap=cmap,
            bins=cmap_bins,
            continuous=continuously,
            colorbar_orientation=colorbar_orientation,
            colorbar_ticks=colorbar_ticks,
            colorbar_ticklabels=colorbar_ticklabels,
            colorbar_label=clabel,
        )
        fig, ax, cax = create_standard_axes(orientation=colorbar_orientation)
        style.colorbar.cax = cax
        if title is None:
            title = default_sweep_title(self.NRadar, 0)
        return plot_vertical_section(
            ax,
            section,
            None,
            None,
            style,
            fig=fig,
            height_km=height,
            title=title,
            labels=axis_labels,
            **kwargs
        )

    def section(
        self,
        start_point,
        end_point,
        field_name,
        range_mode=None,
        height=(0, 18),
        title=None,
        clabel=None,
        continuously=False,
        normvar=None,
        cmap=None,
        cmap_bins=None,
        colorbar_orientation="horizontal",
        colorbar_ticks=None,
        colorbar_ticklabels=None,
        axis_labels=None,
        point_units="km",
        **kwargs
    ):
        ensure_ppi_scan(self.NRadar)
        scale = 1000.0 if point_units == "km" else 1.0
        section = self.NRadar.extract_section(
            (start_point[0] * scale, start_point[1] * scale),
            (end_point[0] * scale, end_point[1] * scale),
            field_name=field_name,
            point_units="m",
            range_mode=range_mode,
        )
        field, field_key = resolve_field_data(self.NRadar, 0, field_name, range_mode=range_mode)
        style = resolve_field_style(
            self.NRadar,
            0,
            field_key,
            field,
            value_range=normvar,
            cmap=cmap,
            bins=cmap_bins,
            continuous=continuously,
            colorbar_orientation=colorbar_orientation,
            colorbar_ticks=colorbar_ticks,
            colorbar_ticklabels=colorbar_ticklabels,
            colorbar_label=clabel,
        )
        fig, ax, cax = create_standard_axes(orientation=colorbar_orientation)
        style.colorbar.cax = cax
        if title is None:
            title = default_sweep_title(self.NRadar, 0)
        return plot_vertical_section(
            ax,
            section,
            None,
            None,
            style,
            fig=fig,
            height_km=height,
            title=title,
            labels=axis_labels,
            **kwargs
        )

    def section_map(
        self,
        start_lonlat,
        end_lonlat,
        field_name,
        range_mode=None,
        height=(0, 18),
        title=None,
        orient="horizontal",
        label=None,
        clabel=None,
        continuously=False,
        normvar=None,
        cmap=None,
        cmap_bins=None,
        colorbar_ticks=None,
        colorbar_ticklabels=None,
        **kwargs
    ):
        ensure_ppi_scan(self.NRadar)
        fig, ax, cax = create_standard_axes(orientation=orient)
        return VerticalSection.GUI_section_map(
            fig,
            ax,
            cax,
            self.NRadar,
            start_lonlat,
            end_lonlat,
            field_name,
            range_mode=range_mode,
            title=title,
            clabel=clabel,
            continuously=continuously,
            normvar=normvar,
            cmap=cmap,
            cmap_bins=cmap_bins,
            colorbar_ticks=colorbar_ticks,
            colorbar_ticklabels=colorbar_ticklabels,
            height=height,
            label=label,
            **kwargs
        )

    @staticmethod
    def GUI_section(
        fig,
        ax,
        cax,
        NRadar,
        start_point,
        end_point,
        field_name,
        range_mode=None,
        title=None,
        clabel=None,
        continuously=False,
        colorbar_orientation="horizontal",
        colorbar_ticks=None,
        colorbar_ticklabels=None,
        normvar=None,
        cmap=None,
        cmap_bins=None,
        height=(0, 18),
        label=None,
        **kwargs
    ):
        ensure_ppi_scan(NRadar)
        section = NRadar.extract_section(
            start_point,
            end_point,
            field_name=field_name,
            point_units="m",
            range_mode=range_mode,
        )
        field, field_key = resolve_field_data(NRadar, 0, field_name, range_mode=range_mode)
        style = resolve_field_style(
            NRadar,
            0,
            field_key,
            field,
            value_range=normvar,
            cmap=cmap,
            bins=cmap_bins,
            continuous=continuously,
            colorbar_orientation=colorbar_orientation,
            colorbar_ticks=colorbar_ticks,
            colorbar_ticklabels=colorbar_ticklabels,
            colorbar_label=clabel,
            colorbar_cax=cax,
        )
        if title is None:
            title = default_sweep_title(NRadar, 0)
        return plot_vertical_section(
            ax,
            section,
            None,
            None,
            style,
            fig=fig,
            height_km=height,
            title=title,
            labels=label,
            **kwargs
        )

    @staticmethod
    def GUI_section_map(
        fig,
        ax,
        cax,
        NRadar,
        start_lonlat,
        end_lonlat,
        field_name,
        range_mode=None,
        title=None,
        clabel=None,
        continuously=False,
        colorbar_ticks=None,
        colorbar_ticklabels=None,
        normvar=None,
        cmap=None,
        cmap_bins=None,
        height=(0, 18),
        label=None,
        orient="horizontal",
        **kwargs
    ):
        ensure_ppi_scan(NRadar)
        field, field_key = resolve_field_data(NRadar, 0, field_name, range_mode=range_mode)
        style = resolve_field_style(
            NRadar,
            0,
            field_key,
            field,
            value_range=normvar,
            cmap=cmap,
            bins=cmap_bins,
            continuous=continuously,
            colorbar_orientation=orient,
            colorbar_ticks=colorbar_ticks,
            colorbar_ticklabels=colorbar_ticklabels,
            colorbar_label=clabel,
            colorbar_cax=cax,
        )
        if title is None:
            title = default_sweep_title(NRadar, 0)
        radar_lon = float(as_numpy(NRadar.scan_info.longitude))
        radar_lat = float(as_numpy(NRadar.scan_info.latitude))
        start_xy, end_xy = geographic_line_to_cartesian(start_lonlat, end_lonlat, radar_lon, radar_lat)
        section = NRadar.extract_section_lonlat(
            start_lonlat,
            end_lonlat,
            field_name=field_name,
            range_mode=range_mode,
        )
        mesh = plot_vertical_section(
            ax,
            section,
            None,
            None,
            style,
            fig=fig,
            height_km=height,
            title=title,
            labels=label,
            **kwargs
        )
        xticks = ax.get_xticks()
        start_xy_km = (start_xy[0] / 1000.0, start_xy[1] / 1000.0)
        end_xy_km = (end_xy[0] / 1000.0, end_xy[1] / 1000.0)
        lon_points, lat_points = lonlat_section_points(start_xy_km, end_xy_km, xticks, radar_lon, radar_lat)
        ax.set_xticklabels(
            ["(%.2f, %.2f)" % (lon_points[i], lat_points[i]) for i in range(len(xticks))],
            rotation=15,
            fontsize=10,
        )
        return mesh

    @staticmethod
    def SectionPlot_VCS(
        fig,
        ax,
        cx,
        mesh_xy,
        mesh_z,
        field_data,
        height=(0, 18),
        title=None,
        normvar=None,
        cmap=None,
        cmap_bins=16,
        orient="horizontal",
        label=None,
        clabel=None,
        continuously=False,
        **kwargs
    ):
        style = PlotStyle(
            cmap=cmap or "viridis",
            value_range=normvar,
            bins=cmap_bins,
            continuous=continuously,
            colorbar=ColorbarOptions(
                visible=True,
                orientation=orient,
                label=clabel,
                cax=cx,
            ),
        )
        return plot_vertical_section(
            ax,
            mesh_xy,
            mesh_z,
            field_data,
            style,
            fig=fig,
            height_km=height,
            title=title,
            labels=label,
            **kwargs
        )

    @staticmethod
    def SectionPlot_VCS_map(
        fig,
        ax,
        cx,
        start_lonlat,
        end_lonlat,
        field_name,
        NRadar,
        range_mode=None,
        height=(0, 18),
        title=None,
        normvar=None,
        cmap=None,
        cmap_bins=16,
        orient="horizontal",
        label=None,
        clabel=None,
        continuously=False,
        **kwargs
    ):
        mesh = VerticalSection.GUI_section_map(
            fig,
            ax,
            cx,
            NRadar,
            start_lonlat,
            end_lonlat,
            field_name,
            range_mode=range_mode,
            title=title,
            clabel=clabel,
            continuously=continuously,
            height=height,
            label=label,
            orient=orient,
            normvar=normvar,
            cmap=cmap,
            cmap_bins=cmap_bins,
            **kwargs
        )
        return mesh

    @staticmethod
    def _FixTicks(ticks):
        ticks = as_numpy(ticks)
        if (ticks % 1).sum() == 0:
            return ["%2.f" % i for i in ticks]
        return ["%.2f" % i for i in ticks]

    @staticmethod
    def get_points_from_ranges(start_points, end_points, distance_from_start):
        return interpolate_points_along_line(start_points, end_points, distance_from_start)



import io
import tempfile
import unittest
from pathlib import Path
from unittest import mock


class ReaderSecurityTests(unittest.TestCase):
    def test_read_all_enforces_maximum_size(self):
        from pycwr.io.util import _read_all

        with self.assertRaises(ValueError):
            _read_all(io.BytesIO(b"0123456789"), "test payload", max_bytes=8)

    def test_wsr98d_reader_uses_larger_payload_cap_for_real_world_archives(self):
        from pycwr.io.WSR98DFile import WSR98DBaseData, WSR98D_MAX_DECODED_PAYLOAD_BYTES

        reader = WSR98DBaseData.__new__(WSR98DBaseData)
        reader.fid = io.BytesIO(b"")

        with mock.patch("pycwr.io.WSR98DFile._read_all", return_value=b"") as read_all:
            radial, status, azimuth, elevation, seconds, microseconds = reader._parse_radial()

        self.assertEqual(radial, [])
        self.assertEqual(status.size, 0)
        self.assertEqual(read_all.call_args.kwargs["max_bytes"], WSR98D_MAX_DECODED_PAYLOAD_BYTES)

    def test_pa_signature_uses_explicit_validation(self):
        from pycwr.io.PAFile import PABaseData

        reader = PABaseData.__new__(PABaseData)
        reader.fid = io.BytesIO(b"\x00" * 12)

        with self.assertRaises(ValueError):
            reader._check_standard_basedata()

    def test_sab_rejects_out_of_bounds_field_offsets(self):
        from pycwr.io.SABFile import SABBaseData

        reader = SABBaseData.__new__(SABBaseData)
        radial_buf = b"\x00" * 2432
        header = {
            "GatesNumberOfReflectivity": 10,
            "PtrOfReflectivity": 4096,
            "GatesNumberOfDoppler": 10,
            "PtrOfVelocity": 0,
            "PtrOfSpectrumWidth": 0,
        }

        with mock.patch("pycwr.io.SABFile._unpack_from_buf", return_value=(header, 128)):
            with self.assertRaises(ValueError):
                reader._parse_radial_single(radial_buf)

    def test_radar_format_returns_none_for_unrelated_archive(self):
        from pycwr.io.util import radar_format

        with tempfile.TemporaryDirectory() as tmpdir:
            archive = Path(tmpdir) / "package.tar.gz"
            archive.write_bytes(b"not-a-radar-archive")
            self.assertIsNone(radar_format(str(archive)))

    def test_radar_format_rejects_false_positive_sab_signature(self):
        from pycwr.io.util import radar_format

        with tempfile.TemporaryDirectory() as tmpdir:
            bogus = Path(tmpdir) / "module.pyc"
            payload = bytearray(b"\x00" * 64)
            payload[14:16] = b"\x01\x00"
            bogus.write_bytes(bytes(payload))
            self.assertIsNone(radar_format(str(bogus)))


class WebViewerSecurityTests(unittest.TestCase):
    def _create_app(self, allowed_root):
        try:
            from pycwr.GraphicalInterface.web_app import create_app
        except ImportError as exc:  # pragma: no cover
            self.skipTest(f"Flask viewer dependencies are unavailable: {exc}")
        return create_app(allowed_roots=[allowed_root], auth_token="test-token")

    def test_api_requires_token(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            app = self._create_app(tmpdir)
            client = app.test_client()
            response = client.get("/api/files", query_string={"dir": tmpdir})
            self.assertEqual(response.status_code, 403)

    def test_api_files_rejects_directory_outside_allowed_roots(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            app = self._create_app(tmpdir)
            client = app.test_client()
            outside_dir = Path(tmpdir).parent
            response = client.get(
                "/api/files",
                query_string={"dir": str(outside_dir), "token": "test-token"},
            )
            self.assertEqual(response.status_code, 403)

    def test_api_metadata_rejects_file_outside_allowed_roots(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            app = self._create_app(tmpdir)
            client = app.test_client()
            outside_file = Path(tmpdir).parent / "outside.bin"
            outside_file.write_bytes(b"not-a-radar-file")
            try:
                response = client.get(
                    "/api/metadata",
                    query_string={"path": str(outside_file), "token": "test-token"},
                )
                self.assertEqual(response.status_code, 403)
            finally:
                outside_file.unlink(missing_ok=True)

    def test_api_files_allows_scanning_within_allowed_roots(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            allowed = Path(tmpdir)
            (allowed / "sample.bin").write_bytes(b"RSTM" + b"\x00" * 128)
            app = self._create_app(tmpdir)
            client = app.test_client()
            response = client.get(
                "/api/files",
                query_string={"dir": str(allowed), "token": "test-token"},
            )
            self.assertEqual(response.status_code, 200)
            payload = response.get_json()
            self.assertTrue(payload["ok"])
            self.assertEqual(len(payload["files"]), 1)
            self.assertEqual(payload["files"][0]["name"], "sample.bin")

    def test_api_tree_recursively_includes_nested_supported_files(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            allowed = Path(tmpdir)
            nested = allowed / "nested" / "day1"
            nested.mkdir(parents=True)
            (nested / "sample.bin").write_bytes(b"RSTM" + b"\x00" * 128)
            app = self._create_app(tmpdir)
            client = app.test_client()
            response = client.get(
                "/api/tree",
                query_string={"dir": str(allowed), "token": "test-token"},
            )

            self.assertEqual(response.status_code, 200)
            payload = response.get_json()
            self.assertTrue(payload["ok"])
            self.assertEqual(payload["tree"]["file_count"], 1)
            nested_node = payload["tree"]["children"][0]["children"][0]["children"][0]
            self.assertEqual(nested_node["type"], "file")
            self.assertEqual(nested_node["name"], "sample.bin")

    def test_api_catalog_groups_files_by_station_and_time(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            allowed = Path(tmpdir)
            (allowed / "Z_RADR_I_Z9046_20260317065928_O_DOR_SAD_CAP_FMT.bin.bz2").write_bytes(b"RSTM" + b"\x00" * 128)
            (allowed / "Z_RADR_I_Z9046_20260317070448_O_DOR_SAD_CAP_FMT.bin.bz2").write_bytes(b"RSTM" + b"\x00" * 128)
            (allowed / "Z_RADR_I_Z9250_20260317065944_O_DOR_SAD_CAP_FMT.bin.bz2").write_bytes(b"RSTM" + b"\x00" * 128)
            app = self._create_app(tmpdir)
            client = app.test_client()
            response = client.get(
                "/api/catalog",
                query_string={"dir": str(allowed), "token": "test-token"},
            )

            self.assertEqual(response.status_code, 200)
            payload = response.get_json()
            self.assertTrue(payload["ok"])
            stations = payload["catalog"]["stations"]
            self.assertEqual([item["station_id"] for item in stations], ["Z9046", "Z9250"])
            self.assertEqual(stations[0]["file_count"], 2)
            self.assertEqual(stations[0]["files"][0]["scan_time"], "2026-03-17T06:59:28")

    def test_default_app_allows_local_directory_scan(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            try:
                from pycwr.GraphicalInterface.web_app import create_app
            except ImportError as exc:  # pragma: no cover
                self.skipTest(f"Flask viewer dependencies are unavailable: {exc}")

            sample = Path(tmpdir) / "sample.bin"
            sample.write_bytes(b"RSTM" + b"\x00" * 128)
            app = create_app(auth_token="test-token")
            client = app.test_client()
            response = client.get(
                "/api/files",
                query_string={"dir": tmpdir, "token": "test-token"},
            )

            self.assertEqual(response.status_code, 200)
            payload = response.get_json()
            self.assertTrue(payload["ok"])
            self.assertEqual(len(payload["files"]), 1)
            self.assertEqual(payload["files"][0]["name"], "sample.bin")

    def test_api_files_requires_supported_filename_suffix(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            allowed = Path(tmpdir)
            sample = allowed / "sample.dat"
            sample.write_bytes(b"RSTM" + b"\x00" * 128)
            app = self._create_app(tmpdir)
            client = app.test_client()
            response = client.get(
                "/api/files",
                query_string={"dir": str(allowed), "token": "test-token"},
            )

            self.assertEqual(response.status_code, 200)
            payload = response.get_json()
            self.assertTrue(payload["ok"])
            self.assertEqual(len(payload["files"]), 0)

    def test_api_tree_ignores_non_radar_archives(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            allowed = Path(tmpdir)
            (allowed / "package.tar.gz").write_bytes(b"not-a-radar-archive")
            app = self._create_app(tmpdir)
            client = app.test_client()
            response = client.get(
                "/api/tree",
                query_string={"dir": str(allowed), "token": "test-token"},
            )

            self.assertEqual(response.status_code, 200)
            payload = response.get_json()
            self.assertTrue(payload["ok"])
            self.assertEqual(payload["tree"]["file_count"], 0)
            self.assertEqual(payload["tree"]["children"], [])

    def test_api_tree_ignores_cache_and_compiled_files(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            allowed = Path(tmpdir)
            cache_dir = allowed / "__pycache__"
            cache_dir.mkdir()
            bogus = cache_dir / "NRadar.cpython-312.pyc"
            payload = bytearray(b"\x00" * 64)
            payload[14:16] = b"\x01\x00"
            bogus.write_bytes(bytes(payload))
            app = self._create_app(tmpdir)
            client = app.test_client()
            response = client.get(
                "/api/tree",
                query_string={"dir": str(allowed), "token": "test-token"},
            )

            self.assertEqual(response.status_code, 200)
            payload = response.get_json()
            self.assertTrue(payload["ok"])
            self.assertEqual(payload["tree"]["file_count"], 0)
            self.assertEqual(payload["tree"]["children"], [])

    def test_plot_section_returns_png(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            allowed = Path(tmpdir)
            sample = allowed / "sample.bin"
            sample.write_bytes(b"placeholder")
            app = self._create_app(tmpdir)
            client = app.test_client()

            from test_examples_sections import SectionExtractionTests

            prd = SectionExtractionTests()._build_ppi_prd()
            with mock.patch("pycwr.GraphicalInterface.web_app.RadarFileCache.get", return_value=prd):
                response = client.get(
                    "/plot/section.png",
                    query_string={
                        "path": str(sample),
                        "token": "test-token",
                        "field": "dBZ",
                        "start_x_km": -10.0,
                        "start_y_km": 0.0,
                        "end_x_km": 10.0,
                        "end_y_km": 0.0,
                    },
                )

            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.mimetype, "image/png")
            self.assertTrue(response.data.startswith(b"\x89PNG"))


if __name__ == "__main__":
    unittest.main()



import importlib
import tempfile
import unittest
from pathlib import Path

import numpy as np


class WSR98DSplitSweepRegressionTests(unittest.TestCase):
    @staticmethod
    def _sample_path():
        return (
            Path(__file__).resolve().parents[2]
            / "Z_RADR_I_ZA202_20250819095753_O_DOR_YLD2-D_CAP_FMT.bin.bz2"
        )

    def _require_pyart(self):
        try:
            return importlib.import_module("pyart")
        except ImportError as exc:  # pragma: no cover
            self.skipTest(f"Py-ART is unavailable: {exc}")

    def test_reader_merges_split_low_level_sweeps(self):
        from pycwr.io import read_auto

        sample = self._sample_path()
        if not sample.exists():
            self.skipTest("ZA202 split-sweep sample is not available in this workspace")

        prd = read_auto(str(sample))
        self.assertEqual(int(prd.nsweeps), 9)
        self.assertEqual(int(prd.nrays), 3290)
        self.assertEqual(prd.resolve_field_name("V", sweep=0), "Vc")
        self.assertEqual(prd.resolve_field_name("W", sweep=0), "W")
        self.assertIn("dBZ", prd.fields[0])
        self.assertIn("Zc", prd.fields[0])
        self.assertIn("Vc", prd.fields[0])
        self.assertIn("W", prd.fields[0])
        self.assertNotIn("V", prd.fields[0])
        self.assertEqual(prd.fields[0]["dBZ"].shape, prd.fields[0]["Vc"].shape)
        self.assertEqual(prd.fields[1]["dBZ"].shape, prd.fields[1]["Vc"].shape)
        self.assertIsNot(prd.fields[0]["dBZ"].variable, prd.fields[0]["Zc"].variable)

    def test_pyart_and_xradar_exports_keep_standard_views_without_merging_storage(self):
        from pycwr.io import read_auto

        sample = self._sample_path()
        if not sample.exists():
            self.skipTest("ZA202 split-sweep sample is not available in this workspace")

        prd = read_auto(str(sample))
        radar = prd.to_pyart_radar(use_external=False, force_rebuild=True)
        sweeps = prd.to_xradar_sweeps(force_rebuild=True)

        self.assertIn("velocity", radar.fields)
        self.assertIn("reflectivity", radar.fields)
        self.assertIn("corrected_reflectivity", radar.fields)
        self.assertIn("sweep_0", sweeps)
        self.assertIn("DBZ", sweeps["sweep_0"].data_vars)
        self.assertIn("VR", sweeps["sweep_0"].data_vars)
        self.assertIn("WRADH", sweeps["sweep_0"].data_vars)

    def test_wsr98d_roundtrip_preserves_corrected_variables(self):
        from pycwr.io import read_WSR98D, read_auto

        sample = self._sample_path()
        if not sample.exists():
            self.skipTest("ZA202 split-sweep sample is not available in this workspace")

        prd = read_auto(str(sample))
        with tempfile.TemporaryDirectory() as tmpdir:
            out = Path(tmpdir) / "za202_roundtrip.bin"
            prd.to_wsr98d(str(out), overwrite=True)
            back = read_WSR98D(str(out))

        self.assertEqual(int(back.nsweeps), int(prd.nsweeps))
        self.assertEqual(int(back.nrays), int(prd.nrays))
        self.assertIn("dBZ", back.available_fields())
        self.assertIn("Zc", back.available_fields())
        self.assertIn("Vc", back.available_fields())
        self.assertNotIn("V", back.available_fields())
        np.testing.assert_allclose(
            np.asarray(back.get_sweep_field(0, "Vc", range_mode=None).values, dtype=np.float64),
            np.asarray(prd.get_sweep_field(0, "Vc", range_mode=None).values, dtype=np.float64),
            atol=0.5 + 1e-6,
            rtol=0.0,
            equal_nan=True,
        )

    def test_nexrad_exports_can_be_read_by_pyart(self):
        pyart = self._require_pyart()
        from pycwr.io import read_auto

        sample = self._sample_path()
        if not sample.exists():
            self.skipTest("ZA202 split-sweep sample is not available in this workspace")

        prd = read_auto(str(sample))
        with tempfile.TemporaryDirectory() as tmpdir:
            msg31 = Path(tmpdir) / "za202_msg31.ar2v"
            msg1 = Path(tmpdir) / "za202_msg1.ar2v"
            prd.to_nexrad_level2_msg31(str(msg31), overwrite=True)
            prd.to_nexrad_level2_msg1(str(msg1), overwrite=True)
            radar31 = pyart.io.read_nexrad_archive(str(msg31))
            radar1 = pyart.io.read_nexrad_archive(str(msg1))

        self.assertEqual(radar31.nsweeps, int(prd.nsweeps))
        self.assertEqual(radar1.nsweeps, int(prd.nsweeps))
        self.assertIn("reflectivity", radar31.fields)
        self.assertIn("velocity", radar31.fields)
        self.assertIn("reflectivity", radar1.fields)
        self.assertIn("velocity", radar1.fields)

    def test_qc_and_network_smoke_on_split_sweep_volume(self):
        from pycwr.core.NRadar import grid_3d_network_xy
        from pycwr.io import read_auto

        sample = self._sample_path()
        if not sample.exists():
            self.skipTest("ZA202 split-sweep sample is not available in this workspace")

        prd = read_auto(str(sample))
        processed = prd.apply_dualpol_qc(inplace=False)
        self.assertIn("Zc", processed.fields[0])
        self.assertIn("ZDRc", processed.fields[0])
        self.assertIn("QC_MASK", processed.fields[0])

        x = np.arange(-20_000.0, 20_001.0, 10_000.0)
        y = np.arange(-20_000.0, 20_001.0, 10_000.0)
        network = grid_3d_network_xy([prd], x, y, np.array([3000.0]), range_mode="native")
        self.assertIn("network_3d", network)
        self.assertEqual(network["network_3d"].shape, (1, x.size, y.size))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Configuration file for the PRD Object , modified from Py-art
注意：距离单位统一到m
"""

FILL_VALUE = -9999.0
_LIGHT_SPEED = 2.99792458e8
HYDROMETEOR_CLASS_NAMES_ZH = (
    "毛毛雨",
    "雨",
    "冰晶",
    "干雪聚合体",
    "湿雪",
    "竖直冰晶",
    "低密度霰",
    "高密度霰",
    "冰雹",
    "大水滴",
)
HYDROMETEOR_CLASS_SHORT_LABELS_ZH = tuple(
    "%d %s" % (index + 1, name) for index, name in enumerate(HYDROMETEOR_CLASS_NAMES_ZH)
)
HYDROMETEOR_CLASS_COLORS = (
    "#9EC5FE",
    "#1F78B4",
    "#D9F0FF",
    "#8FD175",
    "#FFC857",
    "#B39DDB",
    "#F7A35C",
    "#E67E22",
    "#D62828",
    "#FF4FB3",
)

DEFAULT_METADATA = {
    # Metadata for radar attributes. These closely follow the CF/Radial
    # standard

    'azimuth': {
        'units': 'degrees',
        'standard_name': 'beam_azimuth_angle',
        'long_name': 'azimuth_angle_from_true_north',
        'axis': 'radial_azimuth_coordinate',
        'comment': 'Azimuth of antenna relative to true north'},

    'elevation': {
        'units': 'degrees',
        'standard_name': 'beam_elevation_angle',
        'long_name': 'elevation_angle_from_horizontal_plane',
        'axis': 'radial_elevation_coordinate',
        'comment': 'Elevation of antenna relative to the horizontal plane'},

    'range': {
        'units': 'meters',
        'standard_name': 'projection_range_coordinate',
        'long_name': 'range_to_measurement_volume',
        'axis': 'radial_range_coordinate',
        'spacing_is_constant': 'true',
        'comment': (
            'Coordinate variable for range. Range to center of each bin.')},

    'time': {
        'standard_name': 'time',
        'long_name': 'time_in_seconds_since_volume_start',
        'calendar': 'gregorian',
        'comment': ('Coordinate variable for time. '
                    'Time at the center of each ray, in fractional seconds '
                    'since the global variable time_coverage_start')},
    'start_time': {
        'standard_name': 'start_time',
        'long_name': 'UTC_time_at_volume_start',
        'calendar': 'gregorian',
        'comment': ('Coordinate variable for time. '
                    'Time at the center of each ray, in fractional seconds '
                    'since the global variable time_coverage_start')},
    'end_time': {
        'standard_name': 'end_time',
        'long_name': 'UTC_time_at_volume_end',
        'calendar': 'gregorian',
        'comment': ('Coordinate variable for time. '
                    'Time at the center of each ray, in fractional seconds '
                    'since the global variable time_coverage_start')},

    'fixed_angle': {
        'long_name': 'Target angle for sweep',
        'units': 'degrees',
        'standard_name': 'target_fixed_angle'},

    'rays_per_sweep': {
        'long_name': 'Number of rays in each sweep',
        'units': 'count'},

    'latitude': {
        'long_name': 'Latitude',
        'standard_name': 'Latitude',
        'units': 'degrees_north'},
    'lat': {
        'long_name': 'Latitude',
        'standard_name': 'Latitude',
        'units': 'degrees_north'},
    'y': {
        'long_name': 'distance from radar in north',
        'standard_name': 'distance in y',
        'units': 'meters'},

    'longitude': {
        'long_name': 'Longitude',
        'standard_name': 'Longitude',
        'units': 'degrees_east'},

    'lon': {
        'long_name': 'Longitude',
        'standard_name': 'Longitude',
        'units': 'degrees_east'},

    'x': {
        'long_name': 'distance from radar in east',
        'standard_name': 'distance in x',
        'units': 'meters'},

    'altitude': {
        'long_name': 'Altitude',
        'standard_name': 'Altitude',
        'units': 'meters',
        'positive': 'up'},
    'z': {
        'long_name': 'sea surface level',
        'standard_name': 'altitude in z',
        'positive': 'up',
        'units': 'meters'},

    'nyquist_velocity': {
        'units': 'meters_per_second',
        'comments': "Unambiguous velocity",
        'meta_group': 'instrument_parameters',
        'long_name': 'Nyquist velocity'},

    'unambiguous_range': {
        'units': 'meters',
        'comments': 'Unambiguous range',
        'meta_group': 'instrument_parameters',
        'long_name': 'Unambiguous range'},
    'frequency': {
        'units': 'GHZ',
        'meta_group': 'instrument_parameters',
        'long_name': 'Radiation frequency'},

    # Reflectivity fields
    'reflectivity': {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Reflectivity',
        'valid_max': 80.0,
        'valid_min': -30.0,
        'coordinates': 'elevation azimuth range'},
    "corrected_reflectivity": {
        'units': 'dBZ',
        'standard_name': 'corrected_equivalent_reflectivity_factor',
        'long_name': 'Corrected reflectivity',
        'valid_max': 80.0,
        'valid_min': -30.0,
        'coordinates': 'elevation azimuth range'},
    'scan_type':{
        'units': "string",
        'standard_name':"radar scan type",
        'long_name':"Type of scan, one of 'ppi', 'rhi', 'sector' or 'other'"
    },
    'total_power': {
        'units': 'dBZ',
        'standard_name': 'equivalent_reflectivity_factor',
        'long_name': 'Total power',
        'valid_max': 80.0,
        'valid_min': -30.0,
        'coordinates': 'elevation azimuth range'},

    # Velocity fields
    'velocity': {
        'units': 'meters_per_second',
        'standard_name': 'radial_velocity_of_scatterers_away_from_instrument',
        'long_name': 'Mean dopper velocity',
        'valid_max': 50.0,
        'valid_min': -50.0,
        'coordinates': 'elevation azimuth range'},

    "corrected_velocity": {
        'units': 'meters_per_second',
        'standard_name': 'corrected_radial_velocity_of_scatterers_away_from_instrument',
        'valid_max': 50.0,
        'valid_min': -50.0,
        'long_name': 'Corrected mean doppler velocity',
        'coordinates': 'elevation azimuth range'},

    # Spectrum width fields
    'spectrum_width': {
        'units': 'meters_per_second',
        'standard_name': 'doppler_spectrum_width',
        'long_name': 'Doppler spectrum width',
        'valid_max': 30.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},

    'spectrum_width_corrected': {
        'units': 'meters_per_second',
        'standard_name': 'corrected_doppler_spectrum_width',
        'long_name': 'Corrected Doppler spectrum width',
        'valid_max': 30.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},

    # Dual-polarization fields
    'differential_reflectivity': {
        'units': 'dB',
        'standard_name': 'log_differential_reflectivity_hv',
        'long_name': 'Differential reflectivity',
        'valid_max': 8.0,
        'valid_min': -2.0,
        'coordinates': 'elevation azimuth range'},
    'corrected_differential_reflectivity': {
        'units': 'dB',
        'standard_name': 'corrected_log_differential_reflectivity_hv',
        'long_name': 'Corrected Differential reflectivity',
        'valid_max': 8.0,
        'valid_min': -2.0,
        'coordinates': 'elevation azimuth range'},

    'cross_correlation_ratio': {
        'units': 'ratio',
        'standard_name': 'cross_correlation_ratio_hv',
        'long_name': 'Cross correlation ratio (RHOHV)',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},

    'normalized_coherent_power': {
        'units': 'ratio',
        'standard_name': 'normalized_coherent_power',
        'long_name': 'Normalized coherent power',
        'valid_max': 1.0,
        'valid_min': 0.0,
        'comment': 'Also know as signal quality index (SQI)',
        'coordinates': 'elevation azimuth range'},

    'differential_phase': {
        'units': 'degrees',
        'standard_name': 'differential_phase_hv',
        'long_name': 'Differential phase (PhiDP)',
        'valid_max': 360.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},
    'specific_differential_phase': {
        'units': 'degrees/km',
        'standard_name': 'specific_differential_phase_hv',
        'long_name': 'Specific differential phase (KDP)',
        'valid_max': 5,
        'valid_min': -2,
        'coordinates': 'elevation azimuth range'},
    'corrected_specific_differential_phase': {
        'units': 'degrees/km',
        'standard_name': 'corrected_specific_differential_phase_hv',
        'long_name': 'Quality-controlled specific differential phase (KDP)',
        'valid_max': 5,
        'valid_min': -2,
        'coordinates': 'elevation azimuth range'},
    'smoothed_differential_phase': {
        'units': 'degrees',
        'standard_name': 'smoothed_differential_phase_hv',
        'long_name': 'Smoothed differential phase (PhiDP)',
        'valid_max': 360.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},
    'differential_phase_texture': {
        'units': 'degrees',
        'standard_name': 'differential_phase_texture',
        'long_name': 'Differential phase texture',
        'valid_max': 180.0,
        'valid_min': 0.0,
        'coordinates': 'elevation azimuth range'},
    'path_integrated_attenuation': {
        'units': 'dB',
        'standard_name': 'path_integrated_attenuation',
        'long_name': 'Path integrated attenuation',
        'coordinates': 'elevation azimuth range'},
    'path_integrated_differential_attenuation': {
        'units': 'dB',
        'standard_name': 'path_integrated_differential_attenuation',
        'long_name': 'Path integrated differential attenuation',
        'coordinates': 'elevation azimuth range'},
    'meteorological_echo_mask': {
        'units': 'unitless',
        'standard_name': 'meteorological_echo_mask',
        'long_name': 'Meteorological echo mask',
        'flag_values': [0, 1],
        'flag_meanings': 'non_meteorological meteorological',
        'coordinates': 'elevation azimuth range'},
    'quality_control_mask': {
        'units': 'unitless',
        'standard_name': 'quality_control_mask',
        'long_name': 'Quality control mask after despeckling',
        'flag_values': [0, 1],
        'flag_meanings': 'rejected accepted',
        'coordinates': 'elevation azimuth range'},
    'clear_air_echo_mask': {
        'units': 'unitless',
        'standard_name': 'clear_air_echo_mask',
        'long_name': 'Likely clear-air echo mask',
        'flag_values': [0, 1],
        'flag_meanings': 'other_echo likely_clear_air_echo',
        'coordinates': 'elevation azimuth range'},
    'clutter_flag':{
        'standard_name': 'clutter_flag',
    },
    'hydro_class':{
        'standard_name': 'hydro_class',
        'long_name': 'Hydrometeor classification',
        'units': 'class_id',
        'flag_values': list(range(1, len(HYDROMETEOR_CLASS_NAMES_ZH) + 1)),
        'flag_meanings': 'drizzle rain ice_crystals dry_aggregates_snow wet_snow vertical_ice low_density_graupel high_density_graupel hail big_drops',
        'hydrometeor_class_names_zh': ", ".join(HYDROMETEOR_CLASS_NAMES_ZH),
    },
    "horizontal_signal_noise_ratio":{
        'standard_name': 'horizontal signal noise ratio',
    },
    "vertical_signal_noise_ratio":{
        'standard_name': 'vertical signal noise_ratio',
    },
    "flag_of_rpv_data":{
        'standard_name': 'flag of rpv data',
    },
    # Depolarization ratio fields
    'linear_depolarization_ratio': {
        'units': 'dB',
        'standard_name': 'log_linear_depolarization_ratio_hv',
        'long_name': 'Linear depolarization ratio',
        'valid_max': 0,
        'valid_min': -40.0,
        'coordinates': 'elevation azimuth range'},
    "clutter_phase_alignment":{
        'units': 'ratio',
        'standard_name': 'clutter_phase_alignment',
        'long_name': 'clutter phase alignment',
        'valid_max': 0,
        'valid_min': 1,
        'coordinates': 'elevation azimuth range'},
    "beam_width":{
        'units': 'degrees',
        'standard_name': 'clutter_phase_alignment',
        'long_name': 'Antenna beam width polarization',
        'valid_max': 0,
        'valid_min': 5,
        'coordinates': 'sweep'},
}

# CINRAD files
CINRAD_field_mapping = {
    # moment: radar field name
    'dBT':'total_power',
    'dBZ': "reflectivity",
    'V': "velocity",
    'W': "spectrum_width",
    'SQI':'normalized_coherent_power',
    'CPA': 'clutter_phase_alignment',
    'ZDR': "differential_reflectivity",
    'LDR': "linear_depolarization_ratio",
    'CC': "cross_correlation_ratio",
    'PhiDP': "differential_phase",
    'KDP': "specific_differential_phase",
    'KDPc': "corrected_specific_differential_phase",
    'CP': "clutter_probability",
    'Flag': "flag_of_rpv_data",
    'HCL': "hydro_class",
    'CF': "clutter_flag",
    'PhiDP_smooth': "smoothed_differential_phase",
    'PhiDP_texture': "differential_phase_texture",
    'PIA': "path_integrated_attenuation",
    'PIA_ZDR': "path_integrated_differential_attenuation",
    'METEO_MASK': "meteorological_echo_mask",
    'CLEAR_AIR_MASK': "clear_air_echo_mask",
    'QC_MASK': "quality_control_mask",
    'Zc': "corrected_reflectivity",
    'Vc': "corrected_velocity",
    'Wc': "spectrum_width_corrected",
    'SNRH': "horizontal_signal_noise_ratio",
    'SNRV': "vertical_signal_noise_ratio",
    'ZDRc': "corrected_differential_reflectivity"
}

CINRAD_COLORMAP = {
    "reflectivity": 'CN_ref',
    "corrected_reflectivity": 'CN_ref',
    "total_power": 'CN_ref',
    "signal_to_noise_ratio": 'pyart_Carbone17',
    "velocity": 'CN_vel',
    "corrected_velocity": 'pyart_BuDRd18',
    "simulated_velocity": 'pyart_BuDRd18',
    "eastward_wind_component": 'pyart_BuDRd18',
    "northward_wind_component": 'pyart_BuDRd18',
    "vertical_wind_component": 'pyart_BuDRd18',
    "spectrum_width": 'pyart_NWS_SPW',
    "spectrum_width_corrected": 'pyart_NWS_SPW',
    "normalized_coherent_power": 'pyart_Carbone17',
    "differential_reflectivity": 'pyart_RefDiff',
    "corrected_differential_reflectivity": 'pyart_RefDiff',
    "cross_correlation_ratio": 'pyart_RefDiff',
    "differential_phase": 'pyart_RefDiff',
    "unfolded_differential_phase": 'pyart_Wild25',
    "corrected_differential_phase": 'pyart_Wild25',
    "specific_differential_phase": 'pyart_RefDiff',
    "corrected_specific_differential_phase": 'pyart_Theodore16',
    "linear_depolarization_ratio": 'pyart_SCook18',
    "linear_depolarization_ratio_h": 'pyart_SCook18',
    "linear_depolarization_ratio_v": 'pyart_SCook18',
    "rain_rate": 'pyart_RRate11',
    "radar_estimated_rain_rate": 'pyart_RRate11',
    "radar_echo_classification": 'pyart_LangRainbow12',
    "specific_attenuation": 'pyart_Carbone17',
    "differential_phase_texture": 'pyart_BlueBrown11',
    "height": 'pyart_SCook18',
    "interpolated_profile": 'pyart_SCook18',
}

CINRAD_field_bins = {
    # moment: radar field name
    'total_power':16,
    "reflectivity":16,
    "velocity":17,
    "spectrum_width":16,
    "spectrum_width_corrected":16,
    'normalized_coherent_power':16,
    "differential_reflectivity":16,
    "linear_depolarization_ratio":16,
    "cross_correlation_ratio":16,
    "differential_phase":12,
    "specific_differential_phase":12,
    "clutter_flag":16,
    "corrected_reflectivity":16,
    "corrected_velocity":17,
    "signal_to_noise_ratio": 16,
    "simulated_velocity": 16,
    "eastward_wind_component": 16,
    "northward_wind_component": 16,
    "vertical_wind_component": 16,
    "corrected_differential_reflectivity": 16,
    "unfolded_differential_phase": 16,
    "corrected_differential_phase": 16,
    "corrected_specific_differential_phase": 16,
    "linear_depolarization_ratio_h": 16,
    "linear_depolarization_ratio_v": 16,
    "rain_rate": 16,
    "radar_estimated_rain_rate": 16,
    "radar_echo_classification": 16,
    "specific_attenuation": 16,
    "differential_phase_texture": 16,
    "height": 16,
    "interpolated_profile": 16,
}

CINRAD_field_normvar = {
    # moment: radar field name
    'total_power':(-5,75),
    "reflectivity":(-5,75),
    "velocity":(-28,28),
    "spectrum_width":(0,16),
    "spectrum_width_corrected":(0,16),
    'normalized_coherent_power':-1,
    "differential_reflectivity":(-2,8),
    "linear_depolarization_ratio":-1,
    "cross_correlation_ratio":(0,1),
    "differential_phase":(0,180),
    "specific_differential_phase":(-1,6),
    "clutter_flag":-1,
    "corrected_reflectivity":(-5,75),
    "corrected_velocity":(-28, 28),
    "signal_to_noise_ratio": -1,
    "simulated_velocity": -1,
    "eastward_wind_component": -1,
    "northward_wind_component": -1,
    "vertical_wind_component": -1,
    "corrected_differential_reflectivity": -1,
    "unfolded_differential_phase": -1,
    "corrected_differential_phase": -1,
    "corrected_specific_differential_phase": -1,
    "linear_depolarization_ratio_h": -1,
    "linear_depolarization_ratio_v": -1,
    "rain_rate": -1,
    "radar_estimated_rain_rate": -1,
    "radar_echo_classification": -1,
    "specific_attenuation": -1,
    "differential_phase_texture": -1,
    "height": -1,
    "interpolated_profile": -1,
}



# -*- coding: utf-8 -*-
import numpy as np
from .BaseDataProtocol.SCProtocol import dtype_sc
from .util import _prepare_for_read, _read_all, _read_exact, _unpack_from_buf, make_time_unit_str, get_radar_sitename, date2num
import pandas as pd
import datetime
from ..core.NRadar import PRD
from ..configure.pyart_config import get_metadata, get_fillvalue
from ..configure.default_config import CINRAD_field_mapping, _LIGHT_SPEED
from ..core.PyartRadar import Radar

class SCBaseData(object):
    """Decode SC/CD 1.0 base data."""
    def __init__(self, filename, station_lon=None, station_lat=None, station_alt=None):
        """
        :param filename:  radar basedata filename
        :param station_lon:  radar station longitude //units: degree east
        :param station_lat:  radar station latitude //units:degree north
        :param station_alt:  radar station altitude //units: meters
        """
        super(SCBaseData, self).__init__()
        self.filename = filename
        self.station_lon = station_lon
        self.station_lat = station_lat
        self.station_alt = station_alt
        self.fid = _prepare_for_read(self.filename)
        buf_header = _read_exact(self.fid, dtype_sc.BaseDataHeaderSize, "SC header")
        self.header = self._parse_BaseDataHeader(buf_header)
        self.MaxV = self.header['LayerParam']['MaxV'][0] / 100.
        self._radial_buf = self._check_sc_basedata()
        self.radial = self._parse_radial()
        self.fid.close()

    def _check_sc_basedata(self):
        """Check that the radial payload length matches the header metadata."""
        buf_radial_data = _read_all(self.fid, "SC radial payload")
        expected = self.nrays * dtype_sc.PerRadialSize
        if len(buf_radial_data) != expected:
            raise ValueError("SC radial payload size does not match the header metadata.")
        return buf_radial_data

    def _parse_BaseDataHeader(self, buf_header):
        """
        :param buf_header: header-only byte buffer
        :return:
        """
        BaseDataHeader_dict = {}
        # Decode radar site information.
        BaseDataHeader_dict['RadarSite'], _ = _unpack_from_buf(buf_header,\
        dtype_sc.RadarSitePos,dtype_sc.BaseDataHeader['RadarSite'])
        # Decode radar performance parameters.
        BaseDataHeader_dict['RadarPerformanceParam'], _ = _unpack_from_buf(buf_header,\
        dtype_sc.RadarPerformanceParamPos,dtype_sc.BaseDataHeader['RadarPerformanceParam'])
        # Decode the first observation-parameter block.
        BaseDataHeader_dict['RadarObserationParam_1'], _ = _unpack_from_buf(buf_header, \
        dtype_sc.RadarObserationParamPos_1, dtype_sc.BaseDataHeader['RadarObserationParam_1'])
        # Only volume scans are supported here.
        if BaseDataHeader_dict['RadarObserationParam_1']['stype'] <= 100:
            raise ValueError("Only volume-scan SC files are supported.")
        self.nsweeps = BaseDataHeader_dict['RadarObserationParam_1']['stype'] - 100
        # Decode per-sweep layer parameters.
        BaseDataHeader_dict['LayerParam'] = np.frombuffer(buf_header, \
        dtype_sc.BaseDataHeader['LayerParamX30'],count=self.nsweeps, offset=dtype_sc.LayerParamPos)
        # SC base data carries incorrect bin width metadata, keep the historical override.
        BaseDataHeader_dict["binWidth"] = np.full_like(BaseDataHeader_dict['LayerParam']["binWidth"], 5000, dtype=np.int32)
        # Decode the remaining observation parameters.
        BaseDataHeader_dict['RadarObserationParam_2'], _ = _unpack_from_buf(buf_header,\
            dtype_sc.RadarObserationParamPos_2, dtype_sc.BaseDataHeader['RadarObserationParam_2'])
        self.nrays = np.sum((BaseDataHeader_dict['LayerParam']['recordnumber']).astype(np.int64))
        self.sweep_end_ray_index_add1 = np.cumsum((BaseDataHeader_dict['LayerParam']['recordnumber']).astype(np.int64))
        self.sweep_start_ray_index = self.sweep_end_ray_index_add1 - \
                                     (BaseDataHeader_dict['LayerParam']['recordnumber']).astype(np.int64)
        return BaseDataHeader_dict

    def _parse_radial(self):
        radial = []
        payload = self._radial_buf
        pos = 0
        for isweep in range(self.nsweeps):
            MaxV = self.header['LayerParam']['MaxV'][isweep] / 100.
            for _ in range(self.header['LayerParam']['recordnumber'][isweep]):
                buf_radial = payload[pos:pos + dtype_sc.PerRadialSize]
                pos += dtype_sc.PerRadialSize
                radial.append(self._parse_radial_single(buf_radial, 0, MaxV, -1))
        return radial

    def _parse_radial_single(self, buf_radial, start_pos, MaxV, num_bins=-1):
        """
        :param buf_radial:
        :param start_pos: starting byte offset.
        :param num_bins: number of bins.
        :return:
        """
        radial_dict = {}
        radial_dict_tmp, size_tmp = _unpack_from_buf(buf_radial, start_pos, dtype_sc.RadialHeader())
        radial_dict.update(radial_dict_tmp)
        radial_dict['fields'] = {}
        # Historical files may contain invalid trailing radial values.
        #RadialData = np.frombuffer(buf_radial, dtype_sc.RadialData(),\
        #                                 count=num_bins, offset=start_pos+size_tmp)
        RadialData = np.frombuffer(buf_radial, dtype_sc.RadialData(),\
                                         count=500, offset=start_pos+size_tmp)
        radial_dict['fields']['dBZ'] = np.where(RadialData['dBZ'] != 0,\
                            (RadialData['dBZ'].astype(int) - 64)/2., np.nan).astype(np.float32)
        radial_dict['fields']['dBT'] = np.where(RadialData['dBT'] != 0, \
                        (RadialData['dBT'].astype(int) - 64) / 2., np.nan).astype(np.float32)
        radial_dict['fields']['V'] = np.where(RadialData['V'] != 0, \
                       MaxV * (RadialData['V'].astype(int) - 128) / 128., np.nan).astype(np.float32)
        radial_dict['fields']['W'] = np.where(RadialData['W'] != 0, \
                        MaxV * RadialData['W'].astype(int) /256., np.nan).astype(np.float32)
        return radial_dict

    def get_nyquist_velocity(self):
        """Return the per-ray Nyquist velocity."""
        nyquist_velocity = np.concatenate([np.array([self.header['LayerParam']['MaxV'][isweep] / \
                          100.] * self.header['LayerParam']['recordnumber'][isweep]) for \
                            isweep in range(self.nsweeps)])
        return nyquist_velocity.astype(np.float32)
    def get_unambiguous_range(self):
        """Return the per-ray unambiguous range."""
        return np.concatenate([np.array([self.header['LayerParam']['MaxL'][isweep] *10 \
                            ] * self.header['LayerParam']['recordnumber'][isweep]) for \
                            isweep in range(self.nsweeps)])

    def get_scan_time(self):
        """Return the acquisition time for each ray."""
        Start_params = self.header['RadarObserationParam_1']
        End_params = self.header['RadarObserationParam_2']
        start_time = datetime.datetime(year=Start_params['syear'], month=Start_params['smonth'],
                                       day=Start_params['sday'], hour=Start_params['shour'],
                                       minute=Start_params['sminute'], second=Start_params['ssecond'])
        end_time = datetime.datetime(year=End_params['Eyear'], month=End_params['Emonth'],
                                       day=End_params['Eday'], hour=End_params['Ehour'],
                                       minute=End_params['Eminute'], second=End_params['Esecond'])
        return pd.date_range(start_time, end_time, periods=self.nrays).to_pydatetime() - datetime.timedelta(hours=8)

    def get_sweep_end_ray_index(self):
        """Return the inclusive end index of each sweep."""
        return self.sweep_end_ray_index_add1 - 1
    def get_sweep_start_ray_index(self):
        """Return the start index of each sweep."""
        return self.sweep_start_ray_index

    def get_rays_per_sweep(self):
        """Return the number of rays in each sweep."""
        return self.sweep_end_ray_index_add1 - self.sweep_start_ray_index

    def get_azimuth(self):
        """Return the azimuth angle for each ray."""
        return np.concatenate([np.arange(0,360,1.0), ] * self.nsweeps)

    def get_elevation(self):
        """Return the elevation angle for each ray."""
        return np.array([(iray['sStrEl'] + iray['sEndEl'])*180./65536 for iray in self.radial])

    def get_latitude_longitude_altitude_frequency(self):
        """Return latitude, longitude, altitude, and radar frequency."""
        lat, lon, alt, frequency =  self.header['RadarSite']['latitudevalue']/100., \
                                    self.header['RadarSite']['longitudevalue']/100., \
                                    self.header['RadarSite']['height'] / 1000., 2.765
        if self.station_lon is not None:
            lon = self.station_lon
        if self.station_lat is not None:
            lat = self.station_lat
        if self.station_alt is not None:
            alt = self.station_alt
        return lat, lon, alt, frequency

    def get_scan_type(self):
        """Return the scan type string."""
        if self.header['RadarObserationParam_1']['stype'] == 1:
            return "rhi"
        else:
            return "ppi"
    def get_sitename(self):
        return get_radar_sitename(self.filename)

class SC2NRadar(object):
    """Bridge from raw SC data to an NRadar object."""

    def __init__(self, SC):
        self.SC = SC
        self.radial = self.SC.radial
        self.azimuth = self.get_azimuth()
        self.elevation = self.get_elevation()
        self.sweep_start_ray_index = self.get_sweep_start_ray_index()
        self.sweep_end_ray_index = self.get_sweep_end_ray_index()
        self.nrays = self.SC.nrays
        self.nsweeps = self.SC.nsweeps
        self.scan_type = self.SC.get_scan_type()
        self.latitude, self.longitude, self.altitude, self.frequency = \
            self.SC.get_latitude_longitude_altitude_frequency()
        self.bins_per_sweep = self.get_nbins_per_sweep()
        self.max_bins = self.bins_per_sweep.max()
        self.range = self.get_range_per_radial(self.max_bins)
        self.fields = self._get_fields()
        self.sitename = self.SC.get_sitename()

    def get_azimuth(self):
        """Return the azimuth angle for each ray."""
        return self.SC.get_azimuth()

    def get_elevation(self):
        """Return the elevation angle for each ray."""
        return self.SC.get_elevation()

    def get_rays_per_sweep(self):
        """Return the number of rays in each sweep."""
        return (self.SC.header['LayerParam']['recordnumber']).astype(int)

    def get_scan_time(self):
        """Return the acquisition time for each ray."""
        return self.SC.get_scan_time()

    def get_nyquist_velocity(self):
        """Return the per-ray Nyquist velocity."""
        return self.SC.get_nyquist_velocity()

    def get_unambiguous_range(self):
        """Return the per-ray unambiguous range."""
        return self.SC.get_unambiguous_range()

    def get_sweep_end_ray_index(self):
        """Return the inclusive end index of each sweep."""
        return self.SC.sweep_end_ray_index_add1 - 1

    def get_sweep_start_ray_index(self):
        """Return the start index of each sweep."""
        return self.SC.sweep_start_ray_index

    def get_nbins_per_sweep(self):
        """Return the gate count for each sweep."""
        return np.array([self.radial[idx]['fields']['dBZ'].size for idx in self.sweep_start_ray_index])

    def get_range_per_radial(self, length):
        """Return gate-center ranges for a radial of ``length`` bins."""
        Resolution = self.SC.header['binWidth'][0]/10.
        return np.linspace(Resolution, Resolution * length, length)

    def _get_fields(self):
        """Assemble all fields into dense 2-D arrays."""
        fields = {}
        field_keys = self.radial[0]['fields'].keys()
        for ikey in field_keys:
            fields[ikey] = np.array([(iray['fields'][ikey]).ravel() for iray in self.radial])
        return fields

    def get_nradar_nyquist_speed(self):
        """array shape (nsweeps)"""
        return self.SC.header['LayerParam']['MaxV'] / 100.

    def get_NRadar_nyquist_speed(self):
        """Backward-compatible alias for ``get_nradar_nyquist_speed``."""
        return self.get_nradar_nyquist_speed()

    def get_nradar_unambiguous_range(self):
        """array shape (nsweeps)"""
        return self.SC.header['LayerParam']['MaxL'] * 10.

    def get_NRadar_unambiguous_range(self):
        """Backward-compatible alias for ``get_nradar_unambiguous_range``."""
        return self.get_nradar_unambiguous_range()

    def get_fixed_angle(self):
        return self.SC.header['LayerParam']['Swangles'] / 100.

    def to_prd(self, effective_earth_radius=None):
        """Build an ``NRadar.PRD`` object from the decoded SC volume."""

        return PRD(fields=self.fields, scan_type=self.scan_type, time=self.get_scan_time(), \
                          range=self.range, azimuth=self.azimuth, elevation=self.elevation, latitude=self.latitude, \
                          longitude=self.longitude, altitude=self.altitude,
                          sweep_start_ray_index=self.sweep_start_ray_index, \
                          sweep_end_ray_index=self.sweep_end_ray_index, fixed_angle=self.get_fixed_angle(), \
                          bins_per_sweep=self.bins_per_sweep, nyquist_velocity=self.get_nradar_nyquist_speed(), \
                          frequency=self.frequency, unambiguous_range=self.get_nradar_unambiguous_range(), \
                          nrays=self.nrays, nsweeps=self.nsweeps, sitename = self.sitename,
                          pyart_radar=None, effective_earth_radius=effective_earth_radius,
                          metadata={"original_container": "CINRAD/SC", "radar_name": "CINRAD/SC"})

    def ToPRD(self, effective_earth_radius=None):
        """Backward-compatible alias for ``to_prd``."""
        return self.to_prd(effective_earth_radius=effective_earth_radius)

    def to_pyart_radar(self, effective_earth_radius=None, **kwargs):
        """Export the decoded SC volume through the PRD Py-ART adapter."""
        return self.to_prd(effective_earth_radius=effective_earth_radius).to_pyart_radar(**kwargs)

    def ToPyartRadar(self, effective_earth_radius=None, **kwargs):
        """Backward-compatible alias for ``to_pyart_radar``."""
        return self.to_pyart_radar(effective_earth_radius=effective_earth_radius, **kwargs)

    def _get_instrument_parameters(self):
        """ Return a dictionary containing instrument parameters. """

        # pulse width
        pulse_width = get_metadata('pulse_width')
        pulse_width['data'] = np.array([self.range[0] / _LIGHT_SPEED,], dtype='float32')  # nanosec->sec
        # assume that the parameters in the first ray represent the beam widths,
        # bandwidth and frequency in the entire volume
        wavelength_hz = self.frequency * 10 ** 9
        # radar_beam_width_h
        radar_beam_width_h = get_metadata('radar_beam_width_h')
        radar_beam_width_h['data'] = np.array([1., ], dtype='float32')
        # radar_beam_width_v
        radar_beam_width_v = get_metadata('radar_beam_width_v')
        radar_beam_width_v['data'] = np.array([1., ], dtype='float32')
        # frequency
        frequency = get_metadata('frequency')
        frequency['data'] = np.array([wavelength_hz, ], dtype='float32')
        instrument_parameters = {
            'pulse_width': pulse_width,
            'radar_beam_width_h': radar_beam_width_h,
            'radar_beam_width_v': radar_beam_width_v,
            'frequency': frequency, }

        # nyquist velocity if defined
        nyquist_velocity = get_metadata('nyquist_velocity')
        nyquist_velocity['data'] = self.get_nyquist_velocity()
        instrument_parameters['nyquist_velocity'] = nyquist_velocity
        return instrument_parameters



from . import RadarInterp
from .RadarInterp import (
    build_latlon_grid,
    compose_network_volume,
    discover_radar_files,
    load_network_config,
    parse_radar_time_from_filename,
    radar_network_3d_to_netcdf,
    run_radar_network_3d,
    select_radar_files,
)

__all__ = [
    "RadarInterp",
    "build_latlon_grid",
    "compose_network_volume",
    "discover_radar_files",
    "load_network_config",
    "parse_radar_time_from_filename",
    "radar_network_3d_to_netcdf",
    "run_radar_network_3d",
    "select_radar_files",
]



# -*- coding: utf-8 -*-
"""
Attenuation correction helpers.
"""

import numpy as np


KDP_ATTENUATION_COEFFICIENTS = {
    "S": {"gamma": 0.02, "beta": 0.002},
    "C": {"gamma": 0.08, "beta": 0.02},
    "X": {"gamma": 0.28, "beta": 0.04},
}


def _as_float_array(data):
    return np.asanyarray(data, dtype=float)


def _resolve_gate_length_km(data, gate_length):
    if hasattr(data, "range"):
        range_values = np.asanyarray(data.range.values, dtype=float)
        if range_values.size > 1:
            return float(np.median(np.diff(range_values)) / 1000.0)
    return float(gate_length)


def resolve_kdp_coefficients(band="C", gamma=None, beta=None):
    """Return attenuation coefficients for a radar wavelength band."""
    band = str(band).upper()
    if band not in KDP_ATTENUATION_COEFFICIENTS:
        raise ValueError("band must be one of 'S', 'C' or 'X'")
    coefficients = KDP_ATTENUATION_COEFFICIENTS[band]
    if gamma is None:
        gamma = coefficients["gamma"]
    if beta is None:
        beta = coefficients["beta"]
    return float(gamma), float(beta)


def _prepare_mask(data, mask=None):
    valid = np.isfinite(data)
    if mask is None:
        return valid
    return valid & np.asarray(mask, dtype=bool)


def correct_attenuation_HB(Ref, a=1.67e-4, b=0.7, gate_length=0.075, thrs=59):
    """
    Hitschfeld-Bordan attenuation correction.

    Parameters
    ----------
    Ref : array-like
        Reflectivity in dBZ. Range dimension must be the last dimension.
    a, b : float
        Coefficients in the k-Z relation.
    gate_length : float
        Gate length in km. If ``Ref`` is an xarray field exposing ``range``,
        the gate spacing is derived from it automatically.
    thrs : float
        Upper reflectivity threshold. Once exceeded, the current valid segment
        is terminated and the remaining gates in that segment are marked NaN.
    """
    reflectivity = _as_float_array(Ref)
    dr = _resolve_gate_length_km(Ref, gate_length)
    flat_ref = reflectivity.reshape(-1, reflectivity.shape[-1])
    flat_zc = np.full_like(flat_ref, np.nan)
    flat_pia = np.full_like(flat_ref, np.nan)

    for iray, row in enumerate(flat_ref):
        running = 0.0
        overflowed = False
        for igate, value in enumerate(row):
            if not np.isfinite(value):
                running = 0.0
                overflowed = False
                continue
            if overflowed:
                continue
            corrected = value + running
            if corrected > thrs:
                overflowed = True
                continue
            flat_pia[iray, igate] = running
            flat_zc[iray, igate] = corrected
            specific_attenuation = a * (10.0 ** (corrected / 10.0)) ** b
            running += 2.0 * dr * specific_attenuation

    return flat_zc.reshape(reflectivity.shape), flat_pia.reshape(reflectivity.shape)


def correct_attenuation(ref, wavelength="C", rscale=0.075):
    """
    Legacy single-polarization attenuation correction from the historical API.

    This function is retained for backward compatibility. New dual-polarization
    workflows should prefer :func:`correct_attenuation_kdp` or
    :func:`pycwr.qc.run_dualpol_qc`.
    """
    if wavelength == "C":
        a = 0.0044
        b = 1.17
    elif wavelength == "X":
        a = 0.0148
        b = 1.31
    elif wavelength == "S":
        a = 0.0006
        b = 1.00
    else:
        raise ValueError("Unrecognized type")

    if hasattr(ref, "range"):
        rscale = float((ref.range[1] - ref.range[0]) / 1000.0)

    Ra = 200
    Rb = 1.6
    ATT_refl = 4.0
    a_att = 1.0
    b_att = 5.0
    reflectivity = _as_float_array(ref)
    nray, ngate = reflectivity.shape
    Z = np.where(reflectivity > ATT_refl, 10 ** (reflectivity / 10.0), 0.0)
    A = rscale * a * (1.0 / Ra * Z) ** (b / Rb)
    PIA = np.zeros_like(A)
    Asum = np.zeros((nray,))
    for igate in range(1, ngate):
        Asum = Asum + A[:, igate - 1]
        PIA[:, igate] = Asum
    Z_corr = np.where(np.isnan(reflectivity), reflectivity, reflectivity + PIA)
    QI = (b_att - PIA) / (b_att - a_att)
    QI = np.where(PIA < a_att, 1, QI)
    QI = np.where(PIA > b_att, 0, QI)
    return Z_corr, QI


def pia_from_kdp(kdp, dr, gamma=0.08, mask=None, clip_negative=True):
    """
    Retrieve path integrated attenuation from specific differential phase.

    Parameters
    ----------
    kdp : array-like
        Specific differential phase (deg/km). Range dimension must be the last
        dimension.
    dr : float
        Gate length in km.
    gamma : float
        Linear coefficient between KDP and specific attenuation.
    mask : array-like, optional
        Boolean mask indicating gates to include in the integration. Invalid
        gates terminate the current contiguous segment.
    clip_negative : bool
        If True, negative KDP is clipped to zero before integration.
    """
    kdp = _as_float_array(kdp)
    gate_length = float(dr)
    valid = _prepare_mask(kdp, mask=mask)
    flat_kdp = kdp.reshape(-1, kdp.shape[-1])
    flat_valid = valid.reshape(-1, valid.shape[-1])
    flat_pia = np.full_like(flat_kdp, np.nan)

    for iray, (row, row_valid) in enumerate(zip(flat_kdp, flat_valid)):
        running = 0.0
        for igate, (value, is_valid) in enumerate(zip(row, row_valid)):
            if not is_valid:
                running = 0.0
                continue
            if clip_negative:
                value = max(value, 0.0)
            running += 2.0 * gamma * value * gate_length
            flat_pia[iray, igate] = running

    return flat_pia.reshape(kdp.shape)


def correct_attenuation_kdp(ref, kdp, dr, zdr=None, gamma=0.08, beta=0.02, mask=None):
    """
    Dual-polarization attenuation correction driven by KDP.

    Parameters
    ----------
    ref : array-like
        Reflectivity in dBZ.
    kdp : array-like
        Specific differential phase in deg/km.
    dr : float
        Gate length in km.
    zdr : array-like, optional
        Differential reflectivity in dB.
    gamma, beta : float
        Linear coefficients converting KDP to specific attenuation for Z and
        ZDR respectively.
    mask : array-like, optional
        Boolean mask of gates to be corrected.
    """
    reflectivity = _as_float_array(ref)
    kdp = _as_float_array(kdp)
    valid_ref = np.isfinite(reflectivity)
    valid_kdp = np.isfinite(kdp)
    correction_mask = valid_ref & valid_kdp
    if mask is not None:
        correction_mask &= np.asarray(mask, dtype=bool)

    pia = pia_from_kdp(kdp, dr=dr, gamma=gamma, mask=correction_mask, clip_negative=True)
    z_corr = np.where(correction_mask, reflectivity + pia, reflectivity)

    if zdr is None:
        zdr_corr = np.full_like(reflectivity, np.nan)
        pia_zdr = np.full_like(reflectivity, np.nan)
    else:
        differential_reflectivity = _as_float_array(zdr)
        valid_zdr = np.isfinite(differential_reflectivity)
        zdr_mask = correction_mask & valid_zdr
        pia_zdr = pia_from_kdp(kdp, dr=dr, gamma=beta, mask=zdr_mask, clip_negative=True)
        zdr_corr = np.where(zdr_mask, differential_reflectivity + pia_zdr, differential_reflectivity)

    return z_corr, pia, zdr_corr, pia_zdr



# -*- coding: utf-8 -*-
"""Web viewer color palette definitions based on the provided XML scheme."""

from dataclasses import dataclass

import numpy as np
from matplotlib.colors import LinearSegmentedColormap, ListedColormap, PowerNorm

from ..configure.default_config import HYDROMETEOR_CLASS_COLORS, HYDROMETEOR_CLASS_SHORT_LABELS_ZH
from ..draw._plot_core import ColorbarOptions, PlotStyle, default_colorbar_label, resolve_field_style


SPECIAL_COLORS = {
    "Text1": "#ffffff",
    "Text2": "#ffffff",
    "Axis": "#c0c0c0",
    "Ring": "#ffa07a",
    "Symbol": "#ffffff",
    "SymbolUT": "#0000ff",
    "Background1": "#000000",
    "Background2": "#000000",
    "Wind": "#ffffff",
    "WindUT": "#0000ff",
    "RF": "#7c007c",
    "ND": "#7c007c",
}


@dataclass(frozen=True)
class FieldPalette:
    levels: tuple
    labels: tuple
    colors: tuple

    def boundaries(self):
        levels = np.asarray(self.levels, dtype=float)
        if levels.size == 1:
            return np.array([levels[0] - 0.5, levels[0] + 0.5], dtype=float)
        step = levels[-1] - levels[-2]
        if step == 0:
            step = 1.0
        return np.concatenate([levels, [levels[-1] + step]])

    def discrete_cmap(self, field_name):
        return ListedColormap(list(self.colors), name="pycwr_web_%s" % field_name)

    def continuous_cmap(self, field_name):
        return LinearSegmentedColormap.from_list("pycwr_web_linear_%s" % field_name, list(self.colors))


PALETTES = {
    "dBZ": FieldPalette(
        levels=(-5, 0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65),
        labels=("-5", "0", "5", "10", "15", "20", "25", "30", "35", "40", "45", "50", "55", "60", "65"),
        colors=(
            "#00aca4", "#c0c0fe", "#7a72ee", "#1e26d0", "#a6fca8",
            "#00ea00", "#10921a", "#fcf464", "#c8c802", "#8c8c00",
            "#feacac", "#fe6454", "#ee0230", "#d48efe", "#aa24fa",
        ),
    ),
    "dBT": FieldPalette(
        levels=(-5, 0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65),
        labels=("-5", "0", "5", "10", "15", "20", "25", "30", "35", "40", "45", "50", "55", "60", "65"),
        colors=(
            "#00aca4", "#c0c0fe", "#7a72ee", "#1e26d0", "#a6fca8",
            "#00ea00", "#10921a", "#fcf464", "#c8c802", "#8c8c00",
            "#feacac", "#fe6454", "#ee0230", "#d48efe", "#aa24fa",
        ),
    ),
    "V": FieldPalette(
        levels=(-30, -27, -20, -15, -10, -5, -1, 0, 1, 5, 10, 15, 20, 27),
        labels=("-30", "-27", "-20", "-15", "-10", "-5", "-1", "0", "1", "5", "10", "15", "20", "27"),
        colors=(
            "#7ee0fe", "#00e0fe", "#00b0b0", "#00fe00", "#00c400", "#008000",
            "#fefefe", "#fcfcfc", "#fe0000", "#fe5858", "#feb0b0", "#fe7c00",
            "#fed200", "#fefe00",
        ),
    ),
    "W": FieldPalette(
        levels=(0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13),
        labels=("0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12", "13"),
        colors=(
            "#e0e0e0", "#7ce0e0", "#00e0e0", "#00b0b0", "#00fefe", "#00c400",
            "#008000", "#fefe00", "#fed200", "#fe7c00", "#feb0b0", "#af5858",
            "#fe0000", "#e60000",
        ),
    ),
    "SNRH": FieldPalette(
        levels=(-5, 0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65),
        labels=("-5", "0", "5", "10", "15", "20", "25", "30", "35", "40", "45", "50", "55", "60", "65"),
        colors=(
            "#00aca4", "#c0c0fe", "#7a72ee", "#1e26d0", "#a6fca8",
            "#00ea00", "#10921a", "#fcf464", "#c8c802", "#8c8c00",
            "#feacac", "#fe6454", "#ee0230", "#d48efe", "#aa24fa",
        ),
    ),
    "SNRV": FieldPalette(
        levels=(-5, 0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65),
        labels=("-5", "0", "5", "10", "15", "20", "25", "30", "35", "40", "45", "50", "55", "60", "65"),
        colors=(
            "#00aca4", "#c0c0fe", "#7a72ee", "#1e26d0", "#a6fca8",
            "#00ea00", "#10921a", "#fcf464", "#c8c802", "#8c8c00",
            "#feacac", "#fe6454", "#ee0230", "#d48efe", "#aa24fa",
        ),
    ),
    "SQI": FieldPalette(
        levels=(0, 0.06, 0.13, 0.19, 0.25, 0.31, 0.38, 0.44, 0.5, 0.56, 0.62, 0.69, 0.75, 0.81, 0.87, 0.94),
        labels=("0", "0.06", "0.13", "0.19", "0.25", "0.31", "0.38", "0.44", "0.5", "0.56", "0.62", "0.69", "0.75", "0.81", "0.87", "0.94"),
        colors=(
            "#003cff", "#00efef", "#00babf", "#00837d", "#008938", "#00b729",
            "#00da0d", "#00ff00", "#ffff3b", "#fff000", "#ffc600", "#ffa500",
            "#ff7200", "#ff1f00", "#c10000", "#d400aa",
        ),
    ),
    "RR": FieldPalette(
        levels=(0.05, 0.1, 0.2, 0.5, 0.8, 1, 2, 5, 8, 10, 20, 50, 80, 100, 200, 500),
        labels=("0.05", "0.1", "0.2", "0.5", "0.8", "1", "2", "5", "8", "10", "20", "50", "80", "100", "200", "500"),
        colors=(
            "#003cff", "#00efef", "#00babf", "#00837d", "#008938", "#00b729",
            "#00da0d", "#00ff00", "#ffff3b", "#fff000", "#ffc600", "#ffa500",
            "#ff7200", "#ff1f00", "#c10000", "#d400aa",
        ),
    ),
    "ZDR": FieldPalette(
        levels=(-4, -3, -2, -1, 0, 0.2, 0.5, 0.8, 1, 1.5, 2, 2.5, 3, 3.5, 4, 5),
        labels=("-4", "-3", "-2", "-1", "0", "0.2", "0.5", "0.8", "1", "1.5", "2", "2.5", "3", "3.5", "4", "5"),
        colors=(
            "#464646", "#6e6e6e", "#969696", "#c8c8c8", "#dcf0dc", "#00c027",
            "#00e80a", "#24ff24", "#ffff1e", "#ffe600", "#ffbc00", "#ff9800",
            "#ff5e00", "#f20f00", "#bb003a", "#ff00ff",
        ),
    ),
    "LDR": FieldPalette(
        levels=(-30, -28, -26, -24, -22, -20, -18, -16, -14, -12, -10, -8, -6, -4, -2, 0),
        labels=("-30", "-28", "-26", "-24", "-22", "-20", "-18", "-16", "-14", "-12", "-10", "-8", "-6", "-4", "-2", "0"),
        colors=(
            "#003cff", "#00efef", "#00babf", "#00837d", "#008938", "#00b729",
            "#00da0d", "#00ff00", "#ffff3b", "#fff000", "#ffc600", "#ffa500",
            "#ff7200", "#ff1f00", "#c10000", "#d400aa",
        ),
    ),
    "CC": FieldPalette(
        levels=(0, 0.1, 0.3, 0.5, 0.6, 0.7, 0.8, 0.85, 0.9, 0.92, 0.94, 0.95, 0.96, 0.97, 0.98, 0.99, 1.01),
        labels=("0", "0.1", "0.3", "0.5", "0.6", "0.7", "0.8", "0.85", "0.9", "0.92", "0.94", "0.95", "0.96", "0.97", "0.98", "0.99", "1.01"),
        colors=(
            "#003cff", "#00efef", "#00babf", "#00837d", "#008938", "#00b729",
            "#00da0d", "#00ff00", "#ffff3b", "#fff000", "#ffc600", "#ffa500",
            "#ff7200", "#ff1f00", "#c10000", "#d400aa", "#7c007c",
        ),
    ),
    "PhiDP": FieldPalette(
        levels=(0, 24, 48, 72, 96, 120, 144, 168, 192, 216, 240, 264, 288, 312, 336, 360),
        labels=("0", "24", "48", "72", "96", "120", "144", "168", "192", "216", "240", "264", "288", "312", "336", "360"),
        colors=(
            "#003cff", "#00efef", "#00babf", "#00837d", "#008938", "#00b729",
            "#00da0d", "#00ff00", "#ffff3b", "#fff000", "#ffc600", "#ffa500",
            "#ff7200", "#ff1f00", "#c10000", "#d400aa",
        ),
    ),
    "KDP": FieldPalette(
        levels=(-0.8, -0.4, -0.2, -0.1, 0.1, 0.15, 0.22, 0.33, 0.5, 0.75, 1.1, 1.7, 2.4, 3.1, 7, 20),
        labels=("-0.8", "-0.4", "-0.2", "-0.1", "0.1", "0.15", "0.22", "0.33", "0.5", "0.75", "1.1", "1.7", "2.4", "3.1", "7", "20"),
        colors=(
            "#00ffff", "#00efef", "#00a8ac", "#b4b4b4", "#b4b4b4", "#00c027",
            "#00e80a", "#24ff24", "#ffff1e", "#ffe600", "#ffbc00", "#ff9800",
            "#ff5e00", "#f20f00", "#bb003a", "#ff00ff",
        ),
    ),
    "HCL": FieldPalette(
        levels=tuple(range(1, len(HYDROMETEOR_CLASS_SHORT_LABELS_ZH) + 1)),
        labels=HYDROMETEOR_CLASS_SHORT_LABELS_ZH,
        colors=HYDROMETEOR_CLASS_COLORS,
    ),
}


def get_palette(field_name):
    return PALETTES.get(field_name)


def _finite_values(field_data):
    values = np.asarray(getattr(field_data, "values", field_data), dtype=float)
    finite = values[np.isfinite(values)]
    return finite


def _bounded_range(raw_range, lower, upper):
    vmin = max(float(raw_range[0]), float(lower))
    vmax = min(float(raw_range[1]), float(upper))
    if not np.isfinite(vmin) or not np.isfinite(vmax) or vmin >= vmax:
        return float(lower), float(upper)
    return vmin, vmax


def _percentile_range(field_data, low, high, fallback):
    finite = _finite_values(field_data)
    if finite.size == 0:
        return fallback
    vmin = float(np.nanpercentile(finite, low))
    vmax = float(np.nanpercentile(finite, high))
    if not np.isfinite(vmin) or not np.isfinite(vmax) or vmin >= vmax:
        return fallback
    return vmin, vmax


def _preset_range(field_key, field_data, palette, preset):
    palette_range = None
    if palette is not None:
        palette_range = (float(palette.levels[0]), float(palette.levels[-1]))
    if preset == "weak_echo":
        if field_key in ("dBZ", "dBT", "SNRH", "SNRV"):
            return _bounded_range((-5.0, 30.0), *(palette_range or (-5.0, 30.0)))
        if field_key == "RR":
            return _bounded_range((0.05, 50.0), *(palette_range or (0.05, 50.0)))
    detail_range = _percentile_range(field_data, 1.0, 99.0, palette_range or (0.0, 1.0))
    if palette_range is None:
        return detail_range
    return _bounded_range(detail_range, *palette_range)


def build_web_style(
    radar,
    sweep,
    field_key,
    field_data,
    *,
    continuous=False,
    preset="standard",
    clabel=None,
    colorbar_visible=True,
    colorbar_orientation="vertical",
):
    palette = get_palette(field_key)
    preset = (preset or "standard").strip().lower()
    if preset not in ("standard", "fine_detail", "weak_echo"):
        preset = "standard"
    if preset != "standard":
        value_range = _preset_range(field_key, field_data, palette, preset)
        gamma = 0.82 if preset == "fine_detail" else 0.72
        colorbar = ColorbarOptions(
            visible=colorbar_visible,
            orientation=colorbar_orientation,
            label=clabel or default_colorbar_label(field_key),
            ticks=list(palette.levels) if palette is not None else None,
            ticklabels=list(palette.labels) if palette is not None else None,
        )
        if palette is not None:
            return PlotStyle(
                cmap=palette.continuous_cmap(field_key),
                value_range=value_range,
                norm=PowerNorm(gamma=gamma, vmin=value_range[0], vmax=value_range[1], clip=True),
                continuous=True,
                colorbar=colorbar,
            )
        fallback = resolve_field_style(
            radar,
            sweep,
            field_key,
            field_data,
            value_range=value_range,
            continuous=True,
            colorbar_visible=colorbar_visible,
            colorbar_orientation=colorbar_orientation,
            colorbar_label=clabel,
        )
        fallback.norm = PowerNorm(gamma=gamma, vmin=value_range[0], vmax=value_range[1], clip=True)
        return fallback
    if palette is None:
        return resolve_field_style(
            radar,
            sweep,
            field_key,
            field_data,
            continuous=continuous,
            colorbar_visible=colorbar_visible,
            colorbar_orientation=colorbar_orientation,
            colorbar_label=clabel,
        )
    if continuous:
        return PlotStyle(
            cmap=palette.continuous_cmap(field_key),
            value_range=(float(palette.levels[0]), float(palette.levels[-1])),
            continuous=True,
            colorbar=ColorbarOptions(
                visible=colorbar_visible,
                orientation=colorbar_orientation,
                label=clabel or default_colorbar_label(field_key),
                ticks=list(palette.levels),
                ticklabels=list(palette.labels),
            ),
        )
    return PlotStyle(
        cmap=palette.discrete_cmap(field_key),
        levels=palette.boundaries(),
        continuous=False,
        colorbar=ColorbarOptions(
            visible=colorbar_visible,
            orientation=colorbar_orientation,
            label=clabel or default_colorbar_label(field_key),
            ticks=list(palette.levels),
            ticklabels=list(palette.labels),
        ),
    )
